// MODEL.CPP

// Copyright (C) 1998, 2002 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#include "model.h"

#include "filetrans.h"

#include "eng1_qm.h"
#include "eng1_mm.h"
#include "eng2_qm_mm.h"

#ifdef ENABLE_SIMPLIFIED
#include "eng1_sf.h"
#include "eng2_mm_sf.h"
#endif	// ENABLE_SIMPLIFIED

#include "geomopt.h"
#include "moldyn.h"
#include "search.h"

#include <iomanip>
#include <strstream>
using namespace std;

/*################################################################################################*/

char model::local_fn_buffer[256];
sequencebuilder model::amino_builder(chn_info::amino_acid, AMINO_BUILDER_FILE);
sequencebuilder model::nucleic_builder(chn_info::nucleic_acid, NUCLEIC_BUILDER_FILE);

// sf_symbols and sf_types are in the same order.
// st_types are the same as atom::sf_atmtp values.

const char model::sf_symbols[20 + 1] = "ARNDCQEGHILKMFPSTWYV";
const i32s model::sf_types[SF_NUM_TYPES] =
{
	0x0000,				// A
	0x0100, 0x0101, 0x0102,		// R
	0x0200, 0x0201,			// N
	0x0300, 0x0301,			// D
	0x0400, 0x0401,			// C
	0x0500, 0x0501,			// Q
	0x0600, 0x0601,			// E
	0x0700,				// G
	0x0800, 0x0801,			// H
	0x0900, 0x0901,			// I
	0x0A00, 0x0A01,			// L
	0x0B00, 0x0B01, 0x0B02,		// K
	0x0C00, 0x0C01,			// M
	0x0D00, 0x0D01,			// F
	0x0E00,				// P
	0x0F00,				// S
	0x1000,				// T
	0x1100, 0x1101, 0x1102,		// W
	0x1200, 0x1201,			// Y
	0x1300				// V
};

model::model(class_factory & p1)
{
	factory = & p1;
	current_setup = new setup1_mm(this);
	
	err = GetFactory()->ProduceErrUtil();
	model_prefs = GetFactory()->ProducePrefs();
	
	crd_table_size_glob = 1;
	cs_vector.push_back(new crd_set());
//crd_table_size_glob = 2;
//cs_vector.push_back(new crd_set());
//cs_vector.push_back(new crd_set());
//SetCRDSetVisible(1, true);
	SetCRDSetVisible(0, true);
	
	is_index_clean = false;
	is_groups_clean = false;
	
	nmol = NOT_DEFINED;
	ref_civ = NULL;
	
	trajfile = NULL;
	total_traj_frames = NOT_DEFINED;
	current_traj_frame = NOT_DEFINED;
	
	periodic = false;
	box_hdim[0] = 0.5; box_fdim[0] = 2.0 * box_hdim[0];
	box_hdim[1] = 1.0; box_fdim[1] = 2.0 * box_hdim[1];
	box_hdim[2] = 1.5; box_fdim[2] = 2.0 * box_hdim[2];
	
	qm_total_charge = 0;
	qm_current_orbital = 0;
}

model::~model(void)
{
	for (i32u n1 = 0;n1 < cs_vector.size();n1++)
	{
		delete cs_vector[n1];
	}
	
	if (trajfile != NULL) cout << "WARNING - trajectory file was not closed?!?!?!" << endl;
	
	delete err;
}

const char * model::GetProjectFileNameExtension(void)
{
	static const char ext[] = "gpr";
	return ext;
}

const char * model::GetFullPrimaryFN(const char * fn)
{
	ostrstream str(local_fn_buffer, sizeof(local_fn_buffer));
	str << PRIMARY_PATH << VERSION << DIR_SEPARATOR << fn << ends;
	return local_fn_buffer;
}

const char * model::GetFullSecondaryFN(const char * srcdir, const char * fn)
{
	ostrstream str(local_fn_buffer, sizeof(local_fn_buffer));
	str << SECONDARY_PATH << srcdir << fn << ends;
	return local_fn_buffer;
}

void model::OpenParameterFile(ifstream & file, bool is_binary_file, const char * fn)
{
	if (!is_binary_file) file.open(GetFullPrimaryFN(fn), ios::in);
	else file.open(GetFullPrimaryFN(fn), ios::in | ios::binary);
	if (file.good()) return;
	
	file.close();
	cout << "warning: primary path failed for " << GetFullPrimaryFN(fn) << endl;
	
	if (!is_binary_file) file.open(GetFullSecondaryFN("", fn), ios::in);
	else file.open(GetFullSecondaryFN("", fn), ios::in | ios::binary);
	if (file.good()) return;
	
	file.close();
	cout << "fatal error: secondary path failed for " << fn << endl;

	exit(EXIT_FAILURE);
}

void model::UpdateAccumValues(void)		// or NormalizeAccumValues???
{
	fGL sum = 0.0;
	for (i32u n1 = 0;n1 < cs_vector.size();n1++)
	{
		if (GetCRDSetVisible(n1))
		{
			sum += cs_vector[n1]->accum_weight;
		}
	}
	
	for (i32u n1 = 0;n1 < cs_vector.size();n1++)
	{
		if (GetCRDSetVisible(n1))
		{
			fGL tmp1 = cs_vector[n1]->accum_weight / sum;
			cs_vector[n1]->accum_value = tmp1;
		}
	}
}

i32u model::GetCRDSetCount(void)
{
	return cs_vector.size();
}

bool model::GetCRDSetVisible(i32u index)
{
	if (index < cs_vector.size())
	{
		return cs_vector[index]->visible;
	}
	else
	{
		cout << "BUG: invalid call to model::GetCRDSetVisible()" << endl;
		exit(EXIT_FAILURE);
	}
}

void model::SetCRDSetVisible(i32u index, bool visible)
{
	if (index < cs_vector.size())
	{
		cs_vector[index]->visible = visible;
	}
	else
	{
		cout << "BUG: invalid call to model::SetCRDSetVisible()" << endl;
		exit(EXIT_FAILURE);
	}
}

void model::PushCRDSets(i32u p1)
{
	i32u old_size = cs_vector.size();
	
	for (i32u n1 = 0;n1 < p1;n1++) cs_vector.push_back(new crd_set());
	
	i32u new_size = cs_vector.size();
	
	// determine whether we need to reallocate the crd_tables, and reallocate if needed.
	// determine whether we need to reallocate the crd_tables, and reallocate if needed.
	// determine whether we need to reallocate the crd_tables, and reallocate if needed.
	
	if (new_size > crd_table_size_glob)
	{
//		i32u old_cap = crd_table_size_glob;	this is actually not needed...
		i32u new_cap;
		
		new_cap = new_size;	// just allocate the minimum amount (simple).
		
		crd_table_size_glob = new_cap;
		fGL * buff = new fGL[old_size * 3];
		
		for (iter_al it1 = GetAtomsBegin();it1 != GetAtomsEnd();it1++)
		{
			for (i32u n1 = 0;n1 < old_size;n1++)
			{
				buff[n1 * 3 + 0] = (* it1).crd_table[n1 * 3 + 0];
				buff[n1 * 3 + 1] = (* it1).crd_table[n1 * 3 + 1];
				buff[n1 * 3 + 2] = (* it1).crd_table[n1 * 3 + 2];
			}
			
			delete[] (* it1).crd_table;
			(* it1).crd_table = new fGL[new_cap * 3];
			(* it1).crd_table_size_loc = new_cap;
			
			for (i32u n1 = 0;n1 < old_size;n1++)
			{
				(* it1).crd_table[n1 * 3 + 0] = buff[n1 * 3 + 0];
				(* it1).crd_table[n1 * 3 + 1] = buff[n1 * 3 + 1];
				(* it1).crd_table[n1 * 3 + 2] = buff[n1 * 3 + 2];
			}
		}
		
		delete[] buff;
	}
	
	// initialize the new memory blocks.
	// initialize the new memory blocks.
	// initialize the new memory blocks.
	
	for (iter_al it1 = GetAtomsBegin();it1 != GetAtomsEnd();it1++)
	{
		for (i32u n1 = old_size;n1 < new_size;n1++)
		{
			(* it1).crd_table[n1 * 3 + 0] = 0.0;
			(* it1).crd_table[n1 * 3 + 1] = 0.0;
			(* it1).crd_table[n1 * 3 + 2] = 0.0;
		}
	}
}

void model::PopCRDSets(i32u)
{
cout << "Oops!!! This function is not yet ready." << endl;
}

void model::CopyCRDSet(i32u p1, i32u p2)
{
	if (p1 >= crd_table_size_glob || p2 >= crd_table_size_glob)
	{
		cout << "BUG: cs overflow at model::CopyCRDSet()." << endl;
		exit(EXIT_FAILURE);
	}
	
	for (iter_al it1 = atom_list.begin();it1 != atom_list.end();it1++)
	{
		(* it1).crd_table[p2 * 3 + 0] = (* it1).crd_table[p1 * 3 + 0];
		(* it1).crd_table[p2 * 3 + 1] = (* it1).crd_table[p1 * 3 + 1];
		(* it1).crd_table[p2 * 3 + 2] = (* it1).crd_table[p1 * 3 + 2];
	}
}

void model::SwapCRDSets(i32u, i32u)
{
cout << "Oops!!! This function is not yet ready." << endl;
}

void model::CenterCRDSet(i32u p1)
{
	if (p1 >= crd_table_size_glob)
	{
		cout << "BUG: cs overflow at model::CenterCRDSet()." << endl;
		exit(EXIT_FAILURE);
	}
	
	fGL sum[3] = { 0.0, 0.0, 0.0 };
	for (iter_al it1 = atom_list.begin();it1 != atom_list.end();it1++)
	{
		for (i32s n1 = 0;n1 < 3;n1++) sum[n1] += (* it1).crd_table[p1 * 3 + n1];
	}
	
	for (iter_al it1 = atom_list.begin();it1 != atom_list.end();it1++)
	{
		for (i32s n1 = 0;n1 < 3;n1++)
		{
			fGL tmp1 = sum[n1] / (f64) atom_list.size();
			(* it1).crd_table[p1 * 3 + n1] -= tmp1;
		}
	}
}

void model::ReserveCRDSets(i32u)
{
cout << "Oops!!! This function is not yet ready." << endl;
}

void model::DiscardCurrEng(void)
{
//	cout << "discard!!!" << endl;
	GetCurrentSetup()->DiscardCurrentEngine();
}

void model::SetupPlotting(void)
{
	engine * eng = GetCurrentSetup()->GetCurrentEngine();
	if (eng != NULL) eng->SetupPlotting();
}

void model::AddAtom(atom & p1)
{
	DiscardCurrEng();
	InvalidateGroups();
	
	atom_list.push_back(p1);
}

void model::RemoveAtom(iter_al it1)
{
	DiscardCurrEng();
	InvalidateGroups();
	
	// this strange while-loop is needed because removing bonds
	// will invalidate all those atom::cr_list-iterators...
	
	while (!(* it1).cr_list.empty())
	{
		crec * ref = & (* it1).cr_list.back();
		iter_bl it2 = find(bond_list.begin(), bond_list.end(), (* ref->bndr));
		
		if (it2 != bond_list.end()) RemoveBond(it2);
		else
		{
			cout << "BUG: find failed at model::RemoveAtom(); n = " << (* it1).cr_list.size() << endl;
			exit(EXIT_FAILURE);
		}
	}
	
	atom_list.erase(it1);
}

void model::AddBond(bond & p1)
{
	if (p1.atmr[0] == p1.atmr[1])
	{
		cout << "BUG: tried to add an invalid bond at model::AddBond()!" << endl;
		exit(EXIT_FAILURE);
	}
	
	DiscardCurrEng();
	InvalidateGroups();
	
	bond_list.push_back(p1);
	
	crec info1 = crec(p1.atmr[1], & bond_list.back());
	p1.atmr[0]->cr_list.push_back(info1);
/*	if (p1.atmr[0]->cr_list.size() > 8)	// this is for debugging purposes only...
	{
		cout << "BUG: model::AddBond() : too many bonds detected!" << endl;
		exit(EXIT_FAILURE);
	}	*/
	
	crec info2 = crec(p1.atmr[0], & bond_list.back());
	p1.atmr[1]->cr_list.push_back(info2);
/*	if (p1.atmr[1]->cr_list.size() > 8)	// this is for debugging purposes only...
	{
		cout << "BUG: model::AddBond() : too many bonds detected!" << endl;
		exit(EXIT_FAILURE);
	}	*/
}

void model::RemoveBond(iter_bl it1)
{
	DiscardCurrEng();
	InvalidateGroups();
	
	crec tmpinfo = crec(NULL, & (* it1));
	iter_cl it2;
	
	it2 = find((* it1).atmr[0]->cr_list.begin(), (* it1).atmr[0]->cr_list.end(), tmpinfo);
	if (it2 != (* it1).atmr[0]->cr_list.end()) (* it1).atmr[0]->cr_list.erase(it2);
	else
	{
		cout << "BUG: find #1 failed in RemoveBond()" << endl;
		exit(EXIT_FAILURE);
	}
	
	it2 = find((* it1).atmr[1]->cr_list.begin(), (* it1).atmr[1]->cr_list.end(), tmpinfo);
	if (it2 != (* it1).atmr[1]->cr_list.end()) (* it1).atmr[1]->cr_list.erase(it2);
	else
	{
		cout << "BUG: find #2 failed in RemoveBond()" << endl;
		exit(EXIT_FAILURE);
	}
	
	bond_list.erase(it1);
}

/*##############################################*/
/*##############################################*/

bool model::ReadStream(istream & p1, bool selected, bool skipheader)
{
	const int current_version_number = 110;		// version number multiplied by 100 -> int.
	char buffer[1024];
	
	if (!skipheader)
	{
		// first we have to identify the file version, and then call the matching function,
		// if necessary. the very first file format did not use the "!Section" idea, so
		// we must skip the test if no such traces are seen...
		
		int version;
		if (p1.peek() != '!') version = -1;
		else
		{
			// now we expect to see "!Header" section.
			// next we will check the file type and version information...
			
			p1 >> buffer;		// read "!Header".
			p1 >> buffer;		// read the file type.
			p1 >> version;		// read the version number;
			
			p1.getline(buffer, sizeof(buffer));
		}
		
		// now look at the version number, and pick a correct function if needed.
		
		if (version < 110 && version >= 100) return ReadStream_v100(p1, selected, true);
		else if (version < 100) return ReadStream_OLD(p1, selected, true);
		else if (version > current_version_number)
		{
			static const char message[] = "Ooops. It seems that we are trying to read a FUTURE version of this file.\nIt's not sure if we can handle that. Are you sure you wish to try (Y/N) ??? ";
			
			bool result = err->Question(message);
			if (!result) return false;
		}
	}
	
	// the "!Header" section is now processed,
	// and the actual file input begins!!!
	
#ifdef VERBOSE
cout << "model::ReadStream() : starting file input." << endl;
#endif	// VERBOSE

	i32u crd_set_counter = 0;
	vector<atom *> ar_vector;
	vector<bond *> br_vector;
	
	// now we just read the sections, until we hit into "!End"-section and leave the loop...
	
	while (true)
	{
		p1 >> buffer;		// read the section string.
		
		if (!strcmp(buffer, "!End")) break;		// handle "!End" section.
		else if (!strcmp(buffer, "!Info"))		// handle "!Info" section.
		{
			i32s n1; p1 >> n1;		// number of crd-sets
			char cn[16]; p1 >> cn;		// the setyp classname
			i32u engid; p1 >> engid;	// the engine class ID
			
			int n2 = n1 - cs_vector.size();
			if (n2 > 0)
			{
				i32u n3 = cs_vector.size(); PushCRDSets((i32u) n2);
				for (i32u n4 = n3;n4 < cs_vector.size();n4++) SetCRDSetVisible(n4, true);
			}
			
#ifdef VERBOSE
cout << "handled \"!Info\" section, added " << n2 << " coordinate sets." << endl;
cout << "classname = " << cn << " engid = " << engid << " - changing setup to these." << endl;
#endif	// VERBOSE

			p1.getline(buffer, sizeof(buffer));
			
			// TODO : read coordinate set descriptions (if there are some)...
			// TODO : read coordinate set descriptions (if there are some)...
			// TODO : read coordinate set descriptions (if there are some)...
			
			// change the current_setup and current_eng using the settings from file.
			// change the current_setup and current_eng using the settings from file.
			// change the current_setup and current_eng using the settings from file.
			
			delete current_setup;		// get rid of the old setup...
			current_setup = NULL;		// ...and mark it invalid.
			
			// try to find a new setup class; if no success use the setup1_mm as a default.
			
			if (!current_setup && !strcmp(setup1_qm::static_GetClassName(), cn)) current_setup = new setup1_qm(this);
			if (!current_setup && !strcmp(setup1_mm::static_GetClassName(), cn)) current_setup = new setup1_mm(this);
			if (!current_setup && !strcmp(setup2_qm_mm::static_GetClassName(), cn)) current_setup = new setup2_qm_mm(this);
			
#ifdef ENABLE_SIMPLIFIED
			if (!current_setup && !strcmp(setup1_sf::static_GetClassName(), cn)) current_setup = new setup1_sf(this);
			if (!current_setup && !strcmp(setup2_mm_sf::static_GetClassName(), cn)) current_setup = new setup2_mm_sf(this);
#endif	// ENABLE_SIMPLIFIED

			if (!current_setup)
			{
				err->WarningMessage("did not find a matching setup");
				current_setup = new setup1_mm(this);
			}
			
			// now try to set a correct value for current_eng_index
			
			i32u index = 0;
			while (index < GetCurrentSetup()->GetEngineCount())
			{
				i32u id = GetCurrentSetup()->GetEngineIDNumber(index);
				if (id == engid) break;
				
				index++;
			}
			
			if (index == GetCurrentSetup()->GetEngineCount())
			{
				err->WarningMessage("did not find a matching engid");
				index = 0;
			}
			
			GetCurrentSetup()->current_eng_index = index;
		}
		else if (!strcmp(buffer, "!Atoms"))		// handle "!Atoms" section.
		{
			p1.getline(buffer, sizeof(buffer));
			
#ifdef VERBOSE
cout << "found \"!Atoms\" section..." << endl;
#endif	// VERBOSE

			while (p1.peek() != '!')
			{
				i32u n1; p1 >> n1;	// id number
				i32s n2; p1 >> n2;	// atomic number
				i32u n3; p1 >> n3;	// the flags
				
if (n1 != ar_vector.size()) cout << "WARNING : atom id number mismatch!!!" << endl;

				element tmp_element(n2);
				atom newatom(tmp_element, NULL, cs_vector.size());
				
				newatom.flags = n3;
				if (selected) newatom.flags |= ATOMFLAG_SELECTED;
				
				AddAtom(newatom);

				ar_vector.push_back(& atom_list.back());
				
				p1.getline(buffer, sizeof(buffer));
			}
		}
		else if (!strcmp(buffer, "!Bonds"))		// handle "!Bonds" section.
		{
			p1.getline(buffer, sizeof(buffer));
			
#ifdef VERBOSE
cout << "found \"!Bonds\" section..." << endl;
#endif	// VERBOSE

			while (p1.peek() != '!')
			{
				i32u id1; p1 >> id1;	// 1st id number
				i32u id2; p1 >> id2;	// 2nd id number
				
				char bt; p1 >> bt;	// bond type
				
if (id1 >= ar_vector.size()) { cout << "ERROR : invalid atom id number!!!" << endl; exit(EXIT_FAILURE); }
if (id2 >= ar_vector.size()) { cout << "ERROR : invalid atom id number!!!" << endl; exit(EXIT_FAILURE); }

				bondtype tmp_bondtype(bt);
				bond newbond(ar_vector[id1], ar_vector[id2], tmp_bondtype);
				AddBond(newbond);
				
				br_vector.push_back(& bond_list.back());
				
				p1.getline(buffer, sizeof(buffer));
			}
		}
		else if (!strcmp(buffer, "!Coord"))		// handle "!Coord" section.
		{
			p1.getline(buffer, sizeof(buffer));
			
#ifdef VERBOSE
cout << "found \"!Coord\" section..." << endl;
if (crd_set_counter >= cs_vector.size()) { cout << "ERROR : too many crd-sets!!!" << endl; exit(EXIT_FAILURE); }
#endif	// VERBOSE

			while (p1.peek() != '!')
			{
				i32u n1; p1 >> n1;	// id number
				
				fGL xcrd; p1 >> xcrd;	// x-coordinate
				fGL ycrd; p1 >> ycrd;	// y-coordinate
				fGL zcrd; p1 >> zcrd;	// z-coordinate
				
if (n1 >= ar_vector.size()) { cout << "ERROR : invalid atom id number!!!" << endl; exit(EXIT_FAILURE); }

				atom * ar = ar_vector[n1];
				ar->SetCRD(crd_set_counter, xcrd, ycrd, zcrd);
				
				p1.getline(buffer, sizeof(buffer));
			}
			
			crd_set_counter++;
		}
		else if (!strcmp(buffer, "!Charges"))		// handle "!Charges" section.
		{
			p1.getline(buffer, sizeof(buffer));
			
#ifdef VERBOSE
cout << "found \"!Charges\" section..." << endl;
#endif	// VERBOSE

			while (p1.peek() != '!')
			{
				i32u n1; p1 >> n1;	// id number
				f64 n2; p1 >> n2;	// charge
				
if (n1 >= ar_vector.size()) { cout << "ERROR : invalid atom id number!!!" << endl; exit(EXIT_FAILURE); }

				atom * ar = ar_vector[n1];
				ar->charge = n2;
				
				p1.getline(buffer, sizeof(buffer));
			}
		}
		else if (!strcmp(buffer, "!SFAtomInfo"))	// handle the optional "!SFAtomInfo" section.
		{
			p1.getline(buffer, sizeof(buffer));
			
#ifdef VERBOSE
cout << "found \"!SFAtomInfo\" section..." << endl;
#endif	// VERBOSE
			
			while (p1.peek() != '!')
			{
				i32u n1; p1 >> n1;	// id number
				i32u n2; p1 >> n2;	// sf_atmtp
				
if (n1 >= ar_vector.size()) { cout << "ERROR : invalid atom id number!!!" << endl; exit(EXIT_FAILURE); }

				atom * ar = ar_vector[n1];
				if (ar->el.GetAtomicNumber() == -1)
				{
					ar->sf_atmtp = n2;
				}
				else
				{
					cout << "WARNING : an invalid SFAtomInfo record was found; SKIPPING!!!" << endl;
				}
				
				p1.getline(buffer, sizeof(buffer));
			}
		}
		else if (!strcmp(buffer, "!SFBondInfo"))	// handle the optional "!SFBondInfo" section.
		{
			p1.getline(buffer, sizeof(buffer));
			
#ifdef VERBOSE
cout << "found \"!SFBondInfo\" section..." << endl;
#endif	// VERBOSE

			while (p1.peek() != '!')
			{
				i32u n1; p1 >> n1;	// id number
				f64 n2; p1 >> n2;	// sf_pbdd
				
if (n1 >= ar_vector.size()) cout << "ERROR : invalid bond id number!!!" << endl;

				bond * br = br_vector[n1];
				
				bool error = false;
				if (br->atmr[0]->el.GetAtomicNumber() != -1) error = true;	// is it a dummy atom?
				if (br->atmr[1]->el.GetAtomicNumber() != -1) error = true;	// is it a dummy atom?
				if ((br->atmr[0]->sf_atmtp & 0xFF) != 0) error = true;		// is it an alpha carbon?
				if ((br->atmr[1]->sf_atmtp & 0xFF) != 0) error = true;		// is it an alpha carbon?
				
				if (!error)
				{
					br->sf_pbdd = n2;
				}
				else
				{
					cout << "WARNING : an invalid SFBondInfo record was found; SKIPPING!!!" << endl;
				}
				
				p1.getline(buffer, sizeof(buffer));
			}
		}
		else						// handle an unknown section...
		{
			p1.getline(buffer, sizeof(buffer));
			
			while (p1.peek() != '!')
			{
				p1.getline(buffer, sizeof(buffer));
			}
		}
	}
	
	// run some special file integrity checks required by the simplified forcefield.
	// these do not affect any way the usual "all-atoms" files!!! see aa2sf_Convert_sub1().
	
	UpdateIndex();
	for (iter_bl itb = bond_list.begin();itb != bond_list.end();itb++)
	{
		bool flag1a = ((* itb).atmr[0]->el.GetAtomicNumber() == -1);
		bool flag1b = ((* itb).atmr[0]->sf_atmtp != NOT_DEFINED);
		if (!flag1a && !flag1b) continue;
		
		bool flag2a = ((* itb).atmr[1]->el.GetAtomicNumber() == -1);
		bool flag2b = ((* itb).atmr[1]->sf_atmtp != NOT_DEFINED);
		if (!flag2a && !flag2b) continue;
		
		// ok, now the both atoms use simplified presentation.
		cout << "NOTICE : checking SF data integrity for a bond..." << endl;
		
		i32u atnum1 = ((* itb).atmr[0]->sf_atmtp & 0xFF);
		i32u atnum2 = ((* itb).atmr[1]->sf_atmtp & 0xFF);
		
		if (atnum1 == 1 && atnum2 == 1) continue;	// this is probably a valid disulphide bridge...
		
		i32u index1 = (* itb).atmr[0]->index;
		i32u index2 = (* itb).atmr[1]->index;
		
		if (atnum1 == 0 && atnum2 == 0 && (index1 < index2)) continue;		// this is a valid main-chain bond.
		if (atnum1 == 0 && atnum2 == 1 && (index1 < index2)) continue;		// this is a valid side-chain bond.
		if (atnum1 == 1 && atnum2 == 2 && (index1 < index2)) continue;		// this is a valid side-chain bond.
		
		cout << "ERROR : this seems to be a malformed file." << endl;
		exit(EXIT_FAILURE);
	}
	
	// finally, if there were any atoms using simplified presentation, we need to update
	// the vdw-radius and other properties...
	
	sf_UpdateAtomProperties();
	
	// ok, we are ready now!
	
	return true;
}

// WHAT WE NEED TO ADD WHEN WE CHANGE THE FORMAT NEXT TIME:
// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
// [just add your ideas here]

void model::WriteStream(ostream & p1)
{
err->Message("PLEASE NOTE : This file format is experimental!\nIt is safer to use the stable v1.0 of this program.");
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	const int current_version_number = 110;		// version number multiplied by 100 -> int.
	
	// write "!Header" section. this is for file-format definition purposes.
	
	p1 << "!Header "; p1 << "gpr ";
	p1 << current_version_number << endl;
	
	// write "!Info" section. write the number of coordinate sets,
	// which should match to the number of "!Coord"-sections later in the file.
	// also write the ClassName and current EngineID of the setup object that we use now.
// the above only saves the most primitive setup information; need to add !Setup section later???
// the above only saves the most primitive setup information; need to add !Setup section later???
// the above only saves the most primitive setup information; need to add !Setup section later???
	
	p1 << "!Info ";
	p1 << cs_vector.size() << " ";
	p1 << GetCurrentSetup()->GetClassName() << " ";
	p1 << GetCurrentSetup()->GetEngineIDNumber(GetCurrentSetup()->GetCurrEngIndex()) << endl;
	
	// write "!Atoms" section. write the number of atoms,
	// which should match to the number of records in this section.
	
	p1 << "!Atoms ";
	p1 << atom_list.size() << endl;
	
	UpdateIndex();		// this will update the "index"-records...
	
	iter_al it1 = atom_list.begin();
	while (it1 != atom_list.end())
	{
		p1 << (* it1).index << " ";			// unsigned ints.
		p1 << (* it1).el.GetAtomicNumber() << " ";	// signed ints; can be -1 for dummy atoms!!!
		p1 << (* it1).flags << " ";			// unsigned ints; can be large numbers depending on the flags used!
		
		// what else should we print here ?!?!?!?
		
		p1 << endl;
		it1++;
	}
	
	// write "!Bonds" section. write the number of bonds,
	// which should match to the number of records in this section.
	
	p1 << "!Bonds ";
	p1 << bond_list.size() << endl;
	
	iter_bl it2 = bond_list.begin();
	while (it2 != bond_list.end())
	{
		p1 << (* it2).atmr[0]->index << " ";		// unsigned ints.
		p1 << (* it2).atmr[1]->index << " ";		// unsigned ints.
		
		p1 << (* it2).bt.GetSymbol1() << " ";		// single characters.
		
		// what else should we print here ?!?!?!?
		
		p1 << endl;
		it2++;
	}
	
	// write the "!Coord" sections...
// TODO : write visible-flag and set descriptions (if there are some), for example:
// 0 1 "1st coordinate set is this one, and it's visible"
// 1 1 "2nd set looks like this!, and it's also visible"
// 2 0 "this 3rd set is not visible"
	
	for (i32u n1 = 0;n1 < cs_vector.size();n1++)
	{
		p1 << "!Coord" << endl;
		
		it1 = atom_list.begin();
		while (it1 != atom_list.end())
		{
			const fGL * cdata = (* it1).GetCRD(n1);
			
			p1 << (* it1).index << " ";
			p1.setf(ios::showpos);
			p1 << setprecision(15) << cdata[0] << " ";
			p1 << setprecision(15) << cdata[1] << " ";
			p1 << setprecision(15) << cdata[2] << " ";
			p1.unsetf(ios::showpos);
			
			p1 << endl;
			it1++;
		}
	}
	
	// write the !Charges section.
	
	p1 << "!Charges" << endl;
	
	it1 = atom_list.begin();
	while (it1 != atom_list.end())
	{
		p1 << (* it1).index << " ";
		p1.setf(ios::showpos);
		p1 << setprecision(15) << (* it1).charge;
		p1.unsetf(ios::showpos);
		
		p1 << endl;
		it1++;
	}
	
	// optionally, write the !SFAtomInfo section.
	
	i32u nsfatms = 0;
	for (it1 = atom_list.begin();it1 != atom_list.end();it1++)
	{
		if ((* it1).sf_atmtp != NOT_DEFINED) nsfatms++;
	}
	
	if (nsfatms != 0)
	{
		p1 << "!SFAtomInfo" << endl;
		
		it1 = atom_list.begin();
		while (it1 != atom_list.end())
		{
			if ((* it1).sf_atmtp != NOT_DEFINED)
			{
				p1 << (* it1).index << " ";
				p1 << (* it1).sf_atmtp;
				
				p1 << endl;
			}
			
			it1++;
		}
	}
	
	// optionally, write the !SFBondInfo section.
	
	i32u nsfbnds = 0;
	for (it2 = bond_list.begin();it2 != bond_list.end();it2++)
	{
		if ((* it2).sf_pbdd > -1000.0) nsfbnds++;
	}
	
	if (nsfbnds != 0)
	{
		p1 << "!SFBondInfo" << endl;
		
		i32u counter = 0;
		it2 = bond_list.begin();
		while (it2 != bond_list.end())
		{
			if ((* it2).sf_pbdd > -1000.0)
			{
				p1 << counter << " ";
				p1.setf(ios::showpos);
				p1 << setprecision(15) << (* it2).sf_pbdd;
				p1.unsetf(ios::showpos);
				
				p1 << endl;
			}
			
			counter++;
			it2++;
		}
	}
	
	// write "!End" section.
	
	p1 << "!End" << endl;
}

bool model::ImportFile(const char *filename, int index)
{
  ifstream ifile;
  ostrstream intermed;
  file_trans translator;
  
  if (index == 0) // Automatic detection
    {
      if (!translator.CanImport(filename))
	{
	  err->ErrorMessage("Cannot import that file type.");
	  return false;
	}
      
      ifile.open(filename, ios::in);
      translator.Import(filename, ifile, intermed);
      ifile.close();
    }
  else // By type picked by the user
    {
      ifile.open(filename, ios::in);
      translator.Import(filename, index - 1, ifile, intermed);
      ifile.close();
    }

  istrstream interInput(intermed.str());
  return ReadStream(interInput, false);
}

// ExportFile() ???
// ExportFile() ???
// ExportFile() ???

/*##############################################*/
/*##############################################*/

void model::AddHydrogens(void)
{
	file_trans ft;
  	OBMol *obm = ft.CopyAll(this);
	obm->AddHydrogens(false, false);
  	ft.Synchronize();
	
	PrintToLog("Hydrogens added.\n");
}

void model::RemoveHydrogens(void)
{
	iter_bl it1 = bond_list.begin();
	while (it1 != bond_list.end())
	{
		bool flag = false;
		if ((* it1).atmr[0]->el.GetAtomicNumber() == 1) flag = true;
		if ((* it1).atmr[1]->el.GetAtomicNumber() == 1) flag = true;
		iter_bl it2 = it1++; if (flag) RemoveBond(it2);
	}
	
	iter_al it2 = atom_list.begin();
	while (it2 != atom_list.end())
	{
		iter_al it3 = it2++;
		if ((* it3).el.GetAtomicNumber() == 1) RemoveAtom(it3);
	}
	
	PrintToLog("Hydrogens removed.\n");
}

iter_al model::FindAtomByIndex(i32s index)
{
	iter_al end_it = GetAtomsEnd();
	if (index < 0) return end_it;
	
	i32s counter = 0;
	iter_al iter = GetAtomsBegin();
	while (counter != index)
	{
		counter++; iter++;
		if (iter == end_it) return end_it;
	}
	
	return iter;
}

void model::Clear(void)
{
	while (!bond_list.empty())
	{
		iter_bl it1 = bond_list.begin();
		RemoveBond(it1);
	}
	
	while (!atom_list.empty())
	{
		iter_al it1 = atom_list.begin();
		RemoveAtom(it1);
	}
}

/*##############################################*/
/*##############################################*/

void model::GetRange(i32s ind, i32s value, iter_al * result)
{
	iter_al range[2] = { atom_list.begin(), atom_list.end() };
	GetRange(ind, range, value, result);
}

void model::GetRange(i32s ind, iter_al * range, i32s value, iter_al * result)
{
	result[0] = range[0]; while (result[0] != range[1] && (* result[0]).id[ind] != value) result[0]++;
	result[1] = result[0]; while (result[1] != range[1] && (* result[1]).id[ind] == value) result[1]++;
}

i32s model::FindPath(atom * ref1, atom * ref2, i32s max, i32s flag, i32s dist)
{
	if (ref1 == ref2) return dist;
	if (dist == max) return NOT_FOUND;
	
	i32s tmp1 = NOT_FOUND; iter_cl it1;
	for (it1 = ref1->cr_list.begin();it1 != ref1->cr_list.end();it1++)
	{
		if ((* it1).bndr->flags[flag]) continue;
		
		(* it1).bndr->flags[flag] = true;
		i32s tmp2 = FindPath((* it1).atmr, ref2, max, flag, dist + 1);
		(* it1).bndr->flags[flag] = false; if (tmp2 < tmp1) tmp1 = tmp2;
	}
	
	return tmp1;
}

bool model::FindRing(atom * ref1, atom * ref2, signed char * str, i32s size, i32s flag, i32s dist)
{
	static vector<signed char> ring_vector;
		
	if (!dist && str != NULL) ring_vector.resize(0);
	else if (dist && ref1 == ref2)
	{
		if (dist != size) return false;
		
		if (str != NULL)
		{
			for (i32u n1 = 0;n1 < strlen((const char *) str);n1++)
			{
				if (!(n1 % 2) && str[n1] == '?') continue;
				if ((n1 % 2) && str[n1] == NOT_DEFINED) continue;
				if (str[n1] != ring_vector[n1]) return false;
			}
		}
		
		return true;
	}
	
	if (dist == size) return false; iter_cl it1;
	for (it1 = ref1->cr_list.begin();it1 != ref1->cr_list.end();it1++)
	{
		if ((* it1).bndr->flags[flag]) continue;
		
		if (str != NULL)
		{
			ring_vector.push_back((* it1).bndr->bt.GetSymbol2());
			ring_vector.push_back((signed char) (* it1).atmr->el.GetAtomicNumber());
		}
		
		(* it1).bndr->flags[flag] = true;
		bool result = FindRing((* it1).atmr, ref2, str, size, flag, dist + 1);
		(* it1).bndr->flags[flag] = false;
		
		if (str != NULL)
		{
			ring_vector.pop_back();
			ring_vector.pop_back();
		}
		
		if (result) return true;
	}
	
	return false;
}

void model::InvalidateGroups(void)
{
	is_index_clean = false;
	is_groups_clean = false;
	
	nmol = NOT_DEFINED;
	if (ref_civ != NULL)
	{
		delete ref_civ;
		ref_civ = NULL;
	}
}

void model::UpdateIndex(void)
{
	i32s counter = 0;
	iter_al it1 = atom_list.begin();
	while (it1 != atom_list.end())
	{
		(* it1++).index = counter++;
	}
	
	is_index_clean = true;
}

void model::UpdateGroups(void)
{
	InvalidateGroups();
	UpdateIndex();		// looks foolish, but this is a quick operation...
	
	nmol = 0;
	
	iter_al it1;
	for (it1 = atom_list.begin();it1 != atom_list.end();it1++)
	{
		(* it1).id[0] = (* it1).id[1] = (* it1).id[2] = (* it1).id[3] = NOT_DEFINED;
	}
	
	while (true)
	{
		for (it1 = atom_list.begin();it1 != atom_list.end() && (* it1).id[0] != NOT_DEFINED;it1++);
		if (it1 != atom_list.end()) GatherAtoms(& (* it1), nmol++); else break;
	}
	
	atom_list.sort();
	
	is_groups_clean = true;
}

void model::UpdateChains(void)
{
	if (!is_groups_clean) UpdateGroups();
	
	if (ref_civ != NULL) delete ref_civ;
	ref_civ = new vector<chn_info>;
	
	model::amino_builder.Identify(this);
	model::nucleic_builder.Identify(this);
	
	atom_list.sort();
}

void model::GatherAtoms(atom * ref, i32s id)
{
	if (ref->id[0] != NOT_DEFINED) return;
	ref->id[0] = id; iter_cl it1 = ref->cr_list.begin();
	while (it1 != ref->cr_list.end()) GatherAtoms((* it1++).atmr, id);
}

/*##############################################*/
/*##############################################*/

void model::DoEnergy(void)
{
	engine * eng = GetCurrentSetup()->GetCurrentEngine();
	if (eng == NULL) GetCurrentSetup()->CreateCurrentEngine();
	eng = GetCurrentSetup()->GetCurrentEngine();
	if (eng == NULL) return;
	
	char mbuff1[256];
	ostrstream str1(mbuff1, sizeof(mbuff1));
	str1 << "Calculating Energy ";
	str1 << "(setup = " << GetCurrentSetup()->GetClassName();
	str1 << ", engine = " << GetCurrentSetup()->GetEngineName(GetCurrentSetup()->GetCurrEngIndex());
	str1 << ")." << endl << ends;
	PrintToLog(mbuff1);
	
	CopyCRD(this, eng, 0);
	eng->Compute(0);
	
	char buffer[128];
	ostrstream str(buffer, sizeof(buffer)); str.setf(ios::fixed); str.precision(8);
	
	str << "Energy = " << eng->energy << " kJ/mol" << endl << ends;
	PrintToLog(buffer);
	
// we will not delete current_eng here, so that we can draw plots using it...
	
	SetupPlotting();
}

void model::DoGeomOpt(geomopt_param & param)
{
	GeomOptGetParam(param);
	if (!param.confirm) return;
	
	engine * eng = GetCurrentSetup()->GetCurrentEngine();
	if (eng == NULL) GetCurrentSetup()->CreateCurrentEngine();
	eng = GetCurrentSetup()->GetCurrentEngine();
	if (eng == NULL) return;
	
	char mbuff1[256];
	ostrstream str1(mbuff1, sizeof(mbuff1));
	str1 << "Starting Geometry Optimization ";
	str1 << "(setup = " << GetCurrentSetup()->GetClassName();
	str1 << ", engine = " << GetCurrentSetup()->GetEngineName(GetCurrentSetup()->GetCurrEngIndex());
	str1 << ")." << endl << ends;
	PrintToLog(mbuff1);
	
	CopyCRD(this, eng, 0);
	
	geomopt * opt = new geomopt(eng, 100, 0.025);		// optimal settings?!?!?
	
	char buffer[1024];
	f64  last_energy = 0.0;		// this is for output and delta_e test...
	
	PrintToLog("Cycle    Energy    Gradient       Step        Delta E\n");
	
	i32s n1 = 0;	// n1 counts the number of steps...
	while (true)
	{
		opt->TakeCGStep(conjugate_gradient::Newton2An);
		
// problem: the gradient information is in fact not precise in this stage. the current gradient
// is the one that was last calculated in the search, and it is not necessarily the best one.
// to update the gradient, we need to Compute(1) here...
	eng->Compute(1);	// this is not vital, but will update the gradient vector length...
		
		if (n1 != 0)
		  {
		    sprintf(buffer, "%4d %12.5f %10.4e %10.4e %10.4e \n", n1,
			    opt->optval, eng->GetGradientVectorLength(),
			    opt->optstp, last_energy - opt->optval);
		  }
		else
		  {
		    sprintf(buffer, "%4d %12.5f %10.4e %10.4e ********** \n", n1,
			    opt->optval, eng->GetGradientVectorLength(),
			    opt->optstp);
		  }
		PrintToLog(buffer);
		
		bool terminate = false;
		
		if (param.enable_nsteps)	// the nsteps test...
		{
			if (n1 >= param.treshold_nsteps)
			{
				terminate = true;
				PrintToLog("the nsteps termination test was passed.\n");
			}
		}
		
		if (param.enable_grad)		// the grad test...
		{
			if (eng->GetGradientVectorLength() < param.treshold_grad)
			{
				terminate = true;
				PrintToLog("the grad termination test was passed.\n");
			}
		}
		
		if (param.enable_delta_e)	// the delta_e test...
		{
			bool flag = false; const f64 treshold_step = 1.0e-12;		// can we keep this as a constant???
			if (n1 != 0 && (last_energy - opt->optval) != 0.0 && fabs(last_energy - opt->optval) < param.treshold_delta_e) flag = true;
			if ((opt->optstp != 0.0) && (opt->optstp < treshold_step)) flag = true;
			
			if (flag)
			{
				terminate = true;
				PrintToLog("the delta_e termination test was passed.\n");
			}
		}
		
		last_energy = opt->optval;
		
		if (!(n1 % 10) || terminate)
		{
			CopyCRD(eng, this, 0); CenterCRDSet(0);
			UpdateAllGraphicsViews(true);
		}
		
		if (terminate) break;		// exit the loop here!!!
		
		n1++;	// update the number of steps...
	}
	
	delete opt;
	
// we will not delete current_eng here, so that we can draw plots using it...
	
	SetupPlotting();
}

void model::GeomOptGetParam(geomopt_param & param)
{
	param.confirm = true;
}

#define CFRQ 1
#define CSIZE 100

#define START 10000

void model::DoMolDyn(moldyn_param & param)
{
	MolDynGetParam(param);
	if (!param.confirm) return;
	
	engine * eng = GetCurrentSetup()->GetCurrentEngine();
	if (eng == NULL) GetCurrentSetup()->CreateCurrentEngine();
	eng = GetCurrentSetup()->GetCurrentEngine();
	if (eng == NULL) return;
	
//	eng1_mm_default_mim * yyy = NULL;		// temporary...
//	eng1_mm_pbc * zzz = NULL;			// temporary...
	
//	eng1_mm_default_mim * xxx = dynamic_cast<eng1_mm_default_mim *>(eng);
//	if (xxx != NULL) zzz = yyy = xxx;
	
	char mbuff1[256];
	ostrstream str1(mbuff1, sizeof(mbuff1));
	str1 << "Starting Molecular Dynamics ";
	str1 << "(setup = " << GetCurrentSetup()->GetClassName();
	str1 << ", engine = " << GetCurrentSetup()->GetEngineName(GetCurrentSetup()->GetCurrEngIndex());
	str1 << ")." << endl << ends;
	PrintToLog(mbuff1);
	
	CopyCRD(this, eng, 0);
	
	moldyn * dyn = new moldyn(eng, param.temperature, param.timestep);
	
	ofstream ofile;
	ofile.open(param.filename, ios::out | ios::binary);
	
	const char file_id[10] = "traj_v10";
	const int number_of_atoms = atom_list.size();
	const int total_frames = param.nsteps / 100;
	
	ofile.write((char *) file_id, 8);					// file id, 8 chars.
	ofile.write((char *) & number_of_atoms, sizeof(number_of_atoms));	// number of atoms, int.
	ofile.write((char *) & total_frames, sizeof(total_frames));		// total number of frames, int.
	
	i32s ccnt = 0; f64 control[CSIZE];
	
	for (i32s n1 = 0;n1 < START + param.nsteps;n1++)
	{
//		if (!(n1 % 1000) && zzz != NULL) zzz->CheckLocations();
//		if (!(n1 % 100) && yyy != NULL) yyy->update = true;
		
		dyn->TakeMDStep();
		if (!(n1 % 50))
		{
			ostrstream str2(mbuff1, sizeof(mbuff1));
			str2 << "step " << n1 << " T = " << dyn->ConvEKinTemp(dyn->kin);
			str2 << " Epot = " << dyn->pot << " Etot = " << (dyn->kin + dyn->pot) << endl << ends;
			PrintToLog(mbuff1);
		}
		
		if (n1 < START)
		{
			if (!(n1 % CFRQ))
			{
				control[ccnt++] = dyn->kin;
			}
			
			if (ccnt == CSIZE)
			{
				f64 sum = 0.0;
				for (i32s n9 = 0;n9 < ccnt;n9++)
				{
					sum += control[n9];
				}
				
				sum /= (f64) ccnt; ccnt = 0;
				f64 rtemp = (f64) (n1 + CFRQ) / (f64) START;
				
				ostrstream str2(mbuff1, sizeof(mbuff1));
				str2 << "average Ekin = " << sum << " kJ/mol = " << dyn->ConvEKinTemp(sum) << " K." << endl << ends;
				PrintToLog(mbuff1);
				
				dyn->SetEKin(dyn->kin + (dyn->ConvTempEKin(dyn->temp * rtemp) - sum));
			}
		}
		else if (!(n1 % 100))
		{
			CopyCRD(eng, this, 0);
			
			const float ekin = dyn->kin;
			const float epot = dyn->pot;
			
			ofile.write((char *) & ekin, sizeof(ekin));	// kinetic energy, float.
			ofile.write((char *) & epot, sizeof(epot));	// potential energy, float.
			
			for (iter_al it1 = atom_list.begin();it1 != atom_list.end();it1++)
			{
				const fGL * cdata = (* it1).GetCRD(0);
				for (i32s t4 = 0;t4 < 3;t4++)		// all coordinates, float.
				{
					float t1a = cdata[t4];
					ofile.write((char *) & t1a, sizeof(t1a));
				}
			}
		}
		
		if (!(n1 % 100))
		{
			CopyCRD(eng, this, 0);
			UpdateAllGraphicsViews(true);
		}
	}
	
	ofile.close();
	
	delete dyn;
	
	UpdateAccumValues();	// is this needed !??!?!!?
	
// we will not delete current_eng here, so that we can draw plots using it...
	
	SetupPlotting();
}

void model::MolDynGetParam(moldyn_param & param)
{
	param.confirm = true;
}

void model::DoRandomSearch(i32s cycles, i32s optsteps)
{
// eng is not really needed here, but just check that it's available...
// see also project::DoEnergyPlot1D and project::DoEnergyPlot2D
engine * eng = GetCurrentSetup()->GetCurrentEngine();
if (eng == NULL) GetCurrentSetup()->CreateCurrentEngine();
eng = GetCurrentSetup()->GetCurrentEngine();
if (eng == NULL) return;
	
	if (cs_vector.size() < 2)
	{
		PushCRDSets(1);
cout << "PUSH" << endl;
		SetCRDSetVisible(1, false);
	}
	
	random_search rs(this, 0, 0, 1, cycles, optsteps);
	
	while (true)
	{
		int zzz = rs.TakeStep();
	//	cout << zzz << endl;
		
		UpdateAllGraphicsViews(true);
		if (zzz < 0) break;
	}
	
	CopyCRDSet(1, 0);		// get the best structure!!!
	UpdateAllGraphicsViews(true);
	
	char mbuff1[256]; strstream str1(mbuff1, sizeof(mbuff1));
	str1 << "lowest energy found = " << rs.GetMinEnergy() << " kJ/mol" << endl << ends;
	PrintToLog(mbuff1);
	
	PrintToLog("RANDOM SEARCH IS READY.\n");
}

void model::DoSystematicSearch(i32s divisions, i32s optsteps)
{
// eng is not really needed here, but just check that it's available...
// see also project::DoEnergyPlot1D and project::DoEnergyPlot2D
engine * eng = GetCurrentSetup()->GetCurrentEngine();
if (eng == NULL) GetCurrentSetup()->CreateCurrentEngine();
eng = GetCurrentSetup()->GetCurrentEngine();
if (eng == NULL) return;

	if (cs_vector.size() < 2)
	{
		PushCRDSets(1);
		SetCRDSetVisible(1, false);
	}
	
	systematic_search ss(this, 0, 0, 1, divisions, optsteps);
	
	while (true)
	{
		int zzz = ss.TakeStep();
	//	cout << zzz << endl;
		
		UpdateAllGraphicsViews(true);
		if (zzz < 0) break;
	}
	
	CopyCRDSet(1, 0);		// get the best structure!!!
	UpdateAllGraphicsViews(true);
	
	char mbuff1[256]; strstream str1(mbuff1, sizeof(mbuff1));
	str1 << "lowest energy found = " << ss.GetMinEnergy() << " kJ/mol" << endl << ends;
	PrintToLog(mbuff1);
	
	PrintToLog("SYSTEMATIC SEARCH IS READY.\n");
}

void model::DoMonteCarloSearch(i32s n_init_steps, i32s n_simul_steps, i32s optsteps)
{
// eng is not really needed here, but just check that it's available...
// see also project::DoEnergyPlot1D and project::DoEnergyPlot2D
engine * eng = GetCurrentSetup()->GetCurrentEngine();
if (eng == NULL) GetCurrentSetup()->CreateCurrentEngine();
eng = GetCurrentSetup()->GetCurrentEngine();
if (eng == NULL) return;

	
	if (cs_vector.size() < 2)
	{
		PushCRDSets(1);
		SetCRDSetVisible(1, false);
	}
	
	monte_carlo_search mcs(this, 0, 0, 1, n_init_steps, n_simul_steps, optsteps);
	
	while (true)
	{
		int zzz = mcs.TakeStep();
	//	cout << zzz << endl;
		
		UpdateAllGraphicsViews(true);
		if (zzz < 0) break;
	}
	
	CopyCRDSet(1, 0);		// get the best structure!!!
	UpdateAllGraphicsViews(true);
	
	char mbuff1[256]; strstream str1(mbuff1, sizeof(mbuff1));
	str1 << "lowest energy found = " << mcs.GetMinEnergy() << " kJ/mol" << endl << ends;
	PrintToLog(mbuff1);
	
	PrintToLog("MONTE CARLO SEARCH IS READY.\n");
}

void model::DoFormula()
{
  double molweight = 0.0;
  int i;
  char buffer[1024];
  ostrstream str(buffer, sizeof(buffer));

  int count[ELEMENT_SYMBOLS];
  // These are the atomic numbers of the elements in alphabetical order.
  const int alphabetical[ELEMENT_SYMBOLS] = {
   89, 47, 13, 95, 18, 33, 85, 79, 5, 56, 4, 107, 83, 97, 35, 6, 20, 48,
   58, 98, 17, 96, 27, 24, 55, 29, 105, 66, 68, 99, 63, 9, 26, 100, 87, 31,
   64, 32, 1, 2, 72, 80, 67, 108, 53, 49, 77, 19, 36, 57, 3, 103, 71, 101,
   12, 25, 42, 109, 7, 11, 41, 60, 10, 28, 102, 93, 8, 76, 15, 91, 82, 46, 
   61, 84, 59, 78, 94, 88, 37, 75, 104, 45, 86, 44, 16, 51, 21, 34, 106, 14, 
   62, 50, 38, 73, 65, 43, 52, 90, 22, 81, 69, 92, 110, 23, 74, 54, 39, 70, 
   30, 40 };
  int index;

  for (i = 0; i < ELEMENT_SYMBOLS; i++)
    {
      count[i] = 0;
    }

  iter_al it2 = atom_list.begin();
  while (it2 != atom_list.end())
    {
      iter_al it3 = it2++;
      count[(* it3).el.GetAtomicNumber() - 1]++;
      molweight += (* it3).mass;
    }
  
  for (i = 0; i < ELEMENT_SYMBOLS; i++)
    {
      index = alphabetical[i] - 1;
      if (count[index] > 1)
	{
	  str << (element(index + 1)).GetSymbol() << count[index] << " ";
	}
      else if (count[index] == 1)
	{
	  str << (element(index + 1)).GetSymbol();
	}
    }

  str << "MW: " << molweight << ends;
  err->Message(buffer);
}

/*##############################################*/
/*##############################################*/

void model::OpenTrajectory(const char * fn)
{
	if (!trajfile)
	{
		trajfile = new ifstream(fn, ios::in | ios::binary);
		
		trajfile->seekg(8, ios::beg);	// skip the file id...
		
		int natoms;
		trajfile->read((char *) & natoms, sizeof(natoms));
		
		if (natoms != (int) atom_list.size())
		{
			PrintToLog("incompatible file : different number of atoms!\n");
			CloseTrajectory(); return;
		}
		
		trajfile->read((char *) & total_traj_frames, sizeof(total_traj_frames));
		
		char mbuff1[256]; strstream str1(mbuff1, sizeof(mbuff1));
		str1 << "the trajectory file contains " << total_traj_frames << " frames." << endl << ends;
		PrintToLog(mbuff1);
		
		current_traj_frame = 0;
	}
	else PrintToLog("trajectory already exists!\n");
}

void model::CloseTrajectory(void)
{
	if (trajfile != NULL)
	{
		trajfile->close();
		delete trajfile;
		
		trajfile = NULL;
	}
}

void model::ReadFrame(void)
{
	i32s atoms = atom_list.size();
	
	i32s place = 8 + 2 * sizeof(int);					// skip the header...
	place += (2 + 3 * atoms) * sizeof(float) * current_traj_frame;		// get the correct frame...
	place += 2 * sizeof(float);						// skip epot and ekin...
	
	trajfile->seekg(place, ios::beg);
	
	for (iter_al it1 = atom_list.begin();it1 != atom_list.end();it1++)
	{
		fGL cdata[3];
		for (i32s t4 = 0;t4 < 3;t4++)
		{
			float t1a; trajfile->read((char *) & t1a, sizeof(t1a));
			cdata[t4] = t1a;
		}
		
		(* it1).SetCRD(0, cdata[0], cdata[1], cdata[2]);
	}
}

i32s model::GetTotalFrames(void)
{
	return total_traj_frames;
}

ifstream * model::GetTrajectoryFile(void)
{
	return trajfile;
}

i32s model::GetCurrentFrame(void)
{
	return current_traj_frame;
}

void model::SetCurrentFrame(i32s p1)
{
	current_traj_frame = p1;
}

/*##############################################*/
/*##############################################*/

void model::sf_UpdateAtomProperties(void)
{
	for (iter_al it1 = GetAtomsBegin();it1 != GetAtomsEnd();it1++)
	{
		if ((* it1).sf_atmtp != NOT_DEFINED)
		{
			fGL vdwr = -1.0;
			fGL mass = -1.0;
			
			i32s symbol_ind = ((* it1).sf_atmtp >> 8);
			if (symbol_ind < 0 || symbol_ind >= 20) { cout << "BUG: bad symbol_ind." << endl; exit(EXIT_FAILURE); }
			
			i32s atnum = ((* it1).sf_atmtp & 0xFF);
			
			bool error = false;
			switch (sf_symbols[symbol_ind])
			{
				case 'A':								// 71.1
				switch (atnum)
				{
					case 0:
					vdwr = 0.193391;
					mass = 71.1;
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'R':								// 156.7
				switch (atnum)
				{
					case 0:
					vdwr = 0.222079;
					mass = 53.9;			// 1.05
					break;
					
					case 1:
					vdwr = 0.195669;
					mass = 51.4;			// 1.00
					break;
					
					case 2:
					vdwr = 0.189176;
					mass = 51.4;			// 1.00
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'N':								// 114.1
				switch (atnum)
				{
					case 0:
					vdwr = 0.224929;
					mass = 58.4;			// 1.05
					break;
					
					case 1:
					vdwr = 0.187591;
					mass = 55.7;			// 1.00
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'D':								// 114.5
				switch (atnum)
				{
					case 0:
					vdwr = 0.225038;
					mass = 58.6;			// 1.05
					break;
					
					case 1:
					vdwr = 0.187136;
					mass = 55.9;			// 1.00
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'C':								// 103.2
				switch (atnum)
				{
					case 0:
					vdwr = 0.229061;
					mass = 51.6;			// 1.00
					break;
					
					case 1:
					vdwr = 0.179623;
					mass = 51.6;			// 1.00
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'Q':								// 128.2
				switch (atnum)
				{
					case 0:
					vdwr = 0.226117;
					mass = 64.1;			// 1.00
					break;
					
					case 1:
					vdwr = 0.204211;
					mass = 64.1;			// 1.00
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'E':								// 128.7
				switch (atnum)
				{
					case 0:
					vdwr = 0.222040;
					mass = 64.3;			// 1.00
					break;
					
					case 1:
					vdwr = 0.191868;
					mass = 64.3;			// 1.00
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'G':								// 57.1
				switch (atnum)
				{
					case 0:
					vdwr = 0.173764;
					mass = 57.1;
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'H':								// 137.7
				switch (atnum)
				{
					case 0:
					vdwr = 0.229728;
					mass = 68.9;			// 1.00
					break;
					
					case 1:
					vdwr = 0.213733;
					mass = 68.9;			// 1.00
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'I':								// 113.1
				switch (atnum)
				{
					case 0:
					vdwr = 0.224550;
					mass = 56.6;			// 1.00
					break;
					
					case 1:
					vdwr = 0.180242;
					mass = 56.6;			// 1.00
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'L':								// 113.1
				switch (atnum)
				{
					case 0:
					vdwr = 0.220761;
					mass = 56.6;			// 1.00
					break;
					
					case 1:
					vdwr = 0.182951;
					mass = 56.6;			// 1.00
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'K':								// 129.2
				switch (atnum)
				{
					case 0:
					vdwr = 0.222174;
					mass = 49.7;			// 1.25
					break;
					
					case 1:
					vdwr = 0.184924;
					mass = 39.8;			// 1.00
					break;
					
					case 2:
					vdwr = 0.199069;
					mass = 39.8;			// 1.00
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'M':								// 131.2
				switch (atnum)
				{
					case 0:
					vdwr = 0.230562;
					mass = 65.6;			// 1.00
					break;
					
					case 1:
					vdwr = 0.221798;
					mass = 65.6;			// 1.00
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'F':								// 147.2
				switch (atnum)
				{
					case 0:
					vdwr = 0.224414;
					mass = 68.5;			// 1.00
					break;
					
					case 1:
					vdwr = 0.198277;
					mass = 78.7;			// 1.15
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'P':								// 97.1
				switch (atnum)
				{
					case 0:
					vdwr = 0.210783;
					mass = 97.1;
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'S':								// 87.1
				switch (atnum)
				{
					case 0:
					vdwr = 0.204505;
					mass = 87.1;
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'T':								// 101.1
				switch (atnum)
				{
					case 0:
					vdwr = 0.214132;
					mass = 101.1;
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'W':								// 188.2
				switch (atnum)
				{
					case 0:
					vdwr = 0.232409;
					mass = 62.7;			// 1.00
					break;
					
					case 1:
					vdwr = 0.233762;
					mass = 62.7;			// 1.00
					break;
					
					case 2:
					vdwr = 0.225098;
					mass = 62.7;			// 1.00
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'Y':								// 163.2
				switch (atnum)
				{
					case 0:
					vdwr = 0.224633;
					mass = 72.5;			// 1.00
					break;
					
					case 1:
					vdwr = 0.198270;
					mass = 90.6;			// 1.25
					break;
					
					default:
					error = true;
				}
				break;
				
				case 'V':								// 99.1
				switch (atnum)
				{
					case 0:
					vdwr = 0.213718;
					mass = 99.1;
					break;
					
					default:
					error = true;
				}
				break;
				
				default:
				error = true;
			}
			
			if (error)
			{
				cout << "BUG: unknown residue " << sf_symbols[symbol_ind] << atnum << " at sf_UpdateAtomProperties()." << endl;
				exit(EXIT_FAILURE);
			}
			
			(* it1).vdwr = vdwr;
			(* it1).mass = mass;
			
			(* it1).charge = 0.0;
		}
	}
}

void model::aa2sf_ConvertAll(readpdb_mdata * mdata)
{
	model * tmpmdl = aa2sf_Convert_sub1(mdata, false);
	
// do the copying...
// do the copying...
// do the copying...

// no need to worry about readpdb_mdata, or what? of course because it's not even stored here???
	
	while (GetBondCount() != 0) RemoveBond(GetBondsBegin());
	while (GetAtomCount() != 0) RemoveAtom(GetAtomsBegin());
	
	vector<atom *> old_atmr_vector;
	vector<atom *> new_atmr_vector;
	
	for (iter_al it1 = tmpmdl->GetAtomsBegin();it1 != tmpmdl->GetAtomsEnd();it1++)
	{
		const fGL * cdata = (* it1).GetCRD(0);
		atom newatom(element(-1), cdata, GetCRDSetCount());	// element(-1) -> dummy_atom!!!
		
		// "sf_atmtp" and "flags" must be copied from tmpmdl...
		// "sf_atmtp" must be copied from tmpmdl...
		// "sf_atmtp" must be copied from tmpmdl...
		
		newatom.sf_atmtp = (* it1).sf_atmtp;
		newatom.flags = (* it1).flags;
		
		AddAtom(newatom);
		
		old_atmr_vector.push_back(& (* it1));
		new_atmr_vector.push_back(& atom_list.back());
	}
	
	for (iter_bl it1 = tmpmdl->GetBondsBegin();it1 != tmpmdl->GetBondsEnd();it1++)
	{
		i32u ind1 = 0;
		while (ind1 < old_atmr_vector.size())
		{
			if (old_atmr_vector[ind1] == (* it1).atmr[0]) break;
			else ind1++;
		}
		if (ind1 >= old_atmr_vector.size())
		{
			cout << "BUG: aa2sf_ConvertAll() failed #1" << endl;
			exit(EXIT_FAILURE);
		}
		
		i32u ind2 = 0;
		while (ind2 < old_atmr_vector.size())
		{
			if (old_atmr_vector[ind2] == (* it1).atmr[1]) break;
			else ind2++;
		}
		if (ind2 >= old_atmr_vector.size())
		{
			cout << "BUG: aa2sf_ConvertAll() failed #2" << endl;
			exit(EXIT_FAILURE);
		}
		
		bond newbond(new_atmr_vector[ind1], new_atmr_vector[ind2], bondtype('S'));
		
		// "sf_pbdd" must be copied from tmpmdl...
		// "sf_pbdd" must be copied from tmpmdl...
		// "sf_pbdd" must be copied from tmpmdl...
		
		newbond.sf_pbdd = (* it1).sf_pbdd;
		
		AddBond(newbond);
	}
	
	delete tmpmdl;
	
	sf_UpdateAtomProperties();
	InvalidateGroups();
	
#ifdef ENABLE_SIMPLIFIED
	
	// finally, switch the setup class into a suitable one (if available).
	
	delete current_setup; current_setup = new setup1_sf(this);
	err->Message("Ok, the conversion is ready.\nThe setup is also changed to SF.");
	
#else
	
	err->Message("Ok, the conversion is ready.\nNOTE : SF was not enabled in this build;\nrecompile to make it available.");
	
#endif	// ENABLE_SIMPLIFIED

}

void model::aa2sf_ConvertSelection(readpdb_mdata * mdata)
{
	// check that the selection is a continuous block???
	
//model * tmpmdl = aa2sf_Convert_sub1(mdata, true);
	
	// do the above but also check the bonds across boundary!!!
	// and of course retain the not-selected part of the model...
}

model * model::aa2sf_Convert_sub1(readpdb_mdata * mdata, bool)		// bool = check the selection flags???
{
	model * tmpmdl = new model(* console_class_factory::GetInstance());
	tmpmdl->PushCRDSets(GetCRDSetCount() - 1);
	
	// initialize...
	
	UpdateChains();
	DefineSecondaryStructure(this);
	
	// ok, start working...
	
	vector<chn_info> & ci_vector = (* ref_civ);
	
	// now, we must relate mdata and ci_vector!!!
	// for each ci, try to find exactly matching mdata...
	
	i32s * ci_to_mdata_table = new i32s[ci_vector.size()];
	
	for (i32u n1 = 0;n1 < ci_vector.size();n1++)
	{
		ci_to_mdata_table[n1] = NOT_DEFINED;
		if (ci_vector[n1].GetType() != chn_info::amino_acid) continue;
		
		vector<atom *> calpha_vector;
		
		iter_al range1[2]; GetRange(1, n1, range1);
		iter_al range2[2]; i32s res = 0;
		
		while (true)
		{
			GetRange(2, range1, res, range2);
			if (range2[0] == range2[1]) break;
			
			for (iter_al it1 = range2[0];it1 != range2[1];it1++)
			{
				if (((* it1).res_id & 0xFF) != 0x01) continue;
				else calpha_vector.push_back(& (* it1));
			}
			
			res++;		// move to next residue...
		}
		
		if (mdata != NULL)
		{
			for (i32u n2 = 0;n2 < mdata->chn_vector.size();n2++)
			{
				if (calpha_vector.size() != mdata->chn_vector[n2]->alpha_carbons.size()) continue;
				
				bool flag = true;
				for (i32u n3 = 0;n3 < calpha_vector.size();n3++)
				{
					atom * ref1 = calpha_vector[n3];
					atom * ref2 = mdata->chn_vector[n2]->alpha_carbons[n3];
					if (ref1 != ref2) flag = false;
				}
				
				if (flag)
				{
					ci_to_mdata_table[n1] = (i32s) n2;
					break;
				}
			}
			
			if (ci_to_mdata_table[n1] == NOT_DEFINED)
			{
				cout << "failed to map ci " << n1 << " to any mdata record!!!" << endl;
			}
		}
	}
	
	// the main conversion loop.
	
	// also make a simple c-alpha lookup table for the later stages.
	// in the lookup table, the indexation is the same as in ci_vector.
	
	vector< vector<atom *> *> lookup;
	
	vector<i32s> dsb_chn; vector<i32s> dsb_res; vector<atom *> dsb_atm;
	
	fGL * crd_sf[3];
	crd_sf[0] = new fGL[cs_vector.size() * 3];
	crd_sf[1] = new fGL[cs_vector.size() * 3];
	crd_sf[2] = new fGL[cs_vector.size() * 3];
	
	for (i32u n1 = 0;n1 < ci_vector.size();n1++)
	{
		if (ci_vector[n1].GetType() != chn_info::amino_acid)
		{
			lookup.push_back(NULL);
			continue;
		}
		
		vector<atom *> * chn_lookup = new vector<atom *>;
		
		iter_al range1[2];
		GetRange(1, n1, range1);
		
		iter_al range2[2]; i32s res = 0;
		atom * prev_ca = NULL;
		
		while (true)
		{
			GetRange(2, range1, res, range2);
			if (range2[0] == range2[1]) break;
			
			i32s tmp1[2]; tmp1[0] = (* range2[0]).res_id >> 8;
			for (tmp1[1] = 0;tmp1[1] < (i32s) model::amino_builder.residue_vector.size();tmp1[1]++)
			{
				if (model::amino_builder.residue_vector[tmp1[1]].id == tmp1[0]) break;
			} if (tmp1[1] == (i32s) model::amino_builder.residue_vector.size()) continue;
			
			char symbol_sf = model::amino_builder.residue_vector[tmp1[1]].symbol;
			i32s natm_sf = NOT_DEFINED;
			
			switch (symbol_sf)
			{
				case 'A': natm_sf = 1; break;
				
				case 'R': natm_sf = 3; break;
				
				case 'N': natm_sf = 2; break;
				
				case 'D': natm_sf = 2; break;
				
				case 'C': natm_sf = 2; break;
				
				case 'Q': natm_sf = 2; break;
				
				case 'E': natm_sf = 2; break;
				
				case 'G': natm_sf = 1; break;
				
				case 'H': natm_sf = 2; break;
				
				case 'I': natm_sf = 2; break;
				
				case 'L': natm_sf = 2; break;
				
				case 'K': natm_sf = 3; break;
				
				case 'M': natm_sf = 2; break;
				
				case 'F': natm_sf = 2; break;
				
				case 'P': natm_sf = 1; break;
				
				case 'S': natm_sf = 1; break;
				
				case 'T': natm_sf = 1; break;
				
				case 'W': natm_sf = 3; break;
				
				case 'Y': natm_sf = 2; break;
				
				case 'V': natm_sf = 1; break;
				
				default:
				cout << "BUG: unknown residue " << symbol_sf << " at aa2sf_Convert_sub1()." << endl;
				exit(EXIT_FAILURE);
			}
			
			vector<i32s> idv; idv.push_back(0x01);
			for (i32u n2 = 0;n2 < cs_vector.size();n2++)
			{
				aa2sf_Convert_sub2(range2, idv, & crd_sf[0][n2 * 3], n2);
			}
			
			if (natm_sf > 1)
			{
				idv.resize(0);
				switch (symbol_sf)
				{
					case 'R':
					idv.push_back(0x21); idv.push_back(0x22);
					break;
					
					case 'N':
					idv.push_back(0x21);
					break;
					
					case 'D':
					idv.push_back(0x21);
					break;
					
					case 'C':
					idv.push_back(0x21);
					break;
					
					case 'Q':
					idv.push_back(0x21); idv.push_back(0x22);
					break;
					
					case 'E':
					idv.push_back(0x21); idv.push_back(0x22);
					break;
					
					case 'H':
					idv.push_back(0x21); idv.push_back(0x22);
					idv.push_back(0x23); idv.push_back(0x24); idv.push_back(0x25);
					break;
					
					case 'I':
					idv.push_back(0x22);
					break;
					
					case 'L':
					idv.push_back(0x21);
					break;
					
					case 'K':
					idv.push_back(0x21);
					break;
					
					case 'M':
					idv.push_back(0x21); idv.push_back(0x22); idv.push_back(0x23);
					break;
					
					case 'F':
					idv.push_back(0x21); idv.push_back(0x24);
					break;
					
					case 'W':
					idv.push_back(0x21); idv.push_back(0x23); idv.push_back(0x24);
					break;
					
					case 'Y':
					idv.push_back(0x21); idv.push_back(0x24);
					break;
				}
				
				for (i32u n2 = 0;n2 < cs_vector.size();n2++)
				{
					aa2sf_Convert_sub2(range2, idv, & crd_sf[1][n2 * 3], n2);
				}
			}
			
			if (natm_sf > 2)
			{
				idv.resize(0);
				switch (symbol_sf)
				{
					case 'R':
					idv.push_back(0x24);
					break;
					
					case 'K':
					idv.push_back(0x23);
					break;
					
					case 'W':
					idv.push_back(0x25); idv.push_back(0x28);
					break;
				}
				
				for (i32u n2 = 0;n2 < cs_vector.size();n2++)
				{
					aa2sf_Convert_sub2(range2, idv, & crd_sf[2][n2 * 3], n2);
				}
			}
			
			i32s ind = 0; while (ind < 20 && sf_symbols[ind] != symbol_sf) ind++;
			if (ind == 20) { cout << "BUG: unknown residue symbol." << endl; exit(EXIT_FAILURE); }
			
			atom * newres[3];
			for (i32s n2 = 0;n2 < natm_sf;n2++)
			{
				atom newatom(element(-1), & crd_sf[n2][0], tmpmdl->GetCRDSetCount());	// make a dummy atom...
				newatom.sf_atmtp = ((ind << 8) | n2);					// ...and set the sf_atmtp.
				
			//  copy the other crd-sets at this point?!?!?!
			//  copy the other crd-sets at this point?!?!?!
			//  copy the other crd-sets at this point?!?!?!
				
				tmpmdl->AddAtom(newatom);
				newres[n2] = & tmpmdl->atom_list.back();
				
				if (!n2) chn_lookup->push_back(newres[n2]);		// fill the lookup table...
				else							// connect the sidechain atoms...
				{
					bond newbond(newres[n2 - 1], newres[n2], bondtype('S'));	// NOTICE THE ORDER OF ATOMS!!!
					tmpmdl->AddBond(newbond);
				}
				
				if (sf_symbols[ind] == 'C' && n2 == 1)			// this is for disulphide bridges...
				{
					dsb_chn.push_back(n1);
					dsb_res.push_back(((i32s) chn_lookup->size()) - 1);
					dsb_atm.push_back(newres[n2]);
				}
			}
			
			// add bond to the previous residue, if there is one.
			
			if (prev_ca != NULL)
			{
				bond newbond(prev_ca, newres[0], bondtype('S'));			// NOTICE THE ORDER OF ATOMS!!!
				tmpmdl->AddBond(newbond);
			}
			
			prev_ca = newres[0];
			
			res++;		// move to next residue...
		}
		
		lookup.push_back(chn_lookup);
		
		// peptide units...
		
		for (i32s n2 = 1;n2 < ((i32s) chn_lookup->size()) - 2;n2++)
		{
			GetRange(2, range1, n2, range2);
			
			iter_al itc = range2[0];
			while (((* itc).res_id & 0xFF) != 0x01) itc++;
			
			iter_al ito = range2[0];
			while (((* ito).res_id & 0xFF) != 0x10) ito++;
			
			// we only use the crd_set zero...
			
			const fGL * curr = chn_lookup->operator[](n2)->GetCRD(0);
			const fGL * prev = chn_lookup->operator[](n2 - 1)->GetCRD(0);
			const fGL * next = chn_lookup->operator[](n2 + 1)->GetCRD(0);
			
			v3d<fGL> v1(curr, prev); v3d<fGL> v2(curr, next);
			v3d<fGL> v3((* ito).GetCRD(0), (* itc).GetCRD(0));
			fGL pbdd = v1.tor(v2, v3);
			
	bond tmpb1 = bond(chn_lookup->operator[](n2), chn_lookup->operator[](n2 + 1), bondtype());
	iter_bl itb1 = find(tmpmdl->GetBondsBegin(), tmpmdl->GetBondsEnd(), tmpb1);
			
			if (itb1 == tmpmdl->GetBondsEnd())
			{
				cout << "BUG: bond search error." << endl;
				exit(EXIT_FAILURE);
			}
			else (* itb1).sf_pbdd = pbdd;
		}
	}
	
	delete[] crd_sf[0];
	delete[] crd_sf[1];
	delete[] crd_sf[2];
	
	// do some output at this stage (the output should be in FASTA format)...
	
	for (i32u n1 = 0;n1 < lookup.size();n1++)
	{
		cout << "imported chain " << n1 << ":" << endl;
		
		if (ci_to_mdata_table[n1] != NOT_DEFINED)
		{
			i32s index = ci_to_mdata_table[n1];
			i32s chn_length = strlen(mdata->chn_vector[index]->seqres);
			
			cout << "> ORIGINAL sequence, length " << chn_length;
			cout << ", id '" << mdata->chn_vector[index]->chn_id << "':" << endl;
			
			for (i32s n2 = 0;n2 < chn_length;n2++)
			{
				cout << mdata->chn_vector[index]->seqres[n2];
				
				bool is_break = !((n2 + 1) % 50);
				bool is_end = ((n2 + 1) == chn_length);
				
				if (is_break || is_end) cout << endl;
			}
		}
		else
		{
			cout << "> ORIGINAL sequence = ???" << endl;
		}
		
		cout << "> CONVERTED sequence, length ";
		cout << lookup[n1]->size() << ":" << endl;
		
		for (i32u n2 = 0;n2 < lookup[n1]->size();n2++)
		{
			atom * atmr = lookup[n1]->operator[](n2);
			i32s resind = (atmr->sf_atmtp >> 8);
			
			cout << sf_symbols[resind];
			
			bool is_break = !((n2 + 1) % 50);
			bool is_end = ((n2 + 1) == lookup[n1]->size());
			
			if (is_break || is_end) cout << endl;
		}
	}
	
	delete[] ci_to_mdata_table;
	
	// disulphide bridges...
	
	for (i32s n1 = 0;n1 < ((i32s) dsb_atm.size()) - 1;n1++)
	{
		iter_al range1a[2];
		GetRange(1, dsb_chn[n1], range1a);
		
		iter_al range1b[2];
		GetRange(2, range1a, dsb_res[n1], range1b);
		if (range1b[0] == range1b[1]) continue;
		
		iter_al it1 = range1b[0];
		while (it1 != range1b[1] && ((* it1).res_id & 0xFF) != 0x21) it1++;
		if (it1 == range1b[1]) continue;
		
		for (i32s n2 = n1 + 1;n2 < (i32s) dsb_atm.size();n2++)
		{
			iter_al range2a[2];
			GetRange(1, dsb_chn[n2], range2a);
			
			iter_al range2b[2];
			GetRange(2, range2a, dsb_res[n2], range2b);
			if (range2b[0] == range2b[1]) continue;
			
			iter_al it2 = range2b[0];
			while (it2 != range2b[1] && ((* it2).res_id & 0xFF) != 0x21) it2++;
			if (it2 == range2b[1]) continue;
			
			// test if the brigde exists...
			
			bond tmpbond = bond(& (* it1), & (* it2), bondtype('S'));
			iter_bl it1 = find(bond_list.begin(), bond_list.end(), tmpbond);
			if (it1 != bond_list.end())
			{
				cout << "adding disulphide bridge ";
				cout << "c1 = " << dsb_chn[n1] << " r1 = " << dsb_res[n1] << " ";
				cout << "c2 = " << dsb_chn[n2] << " r2 = " << dsb_res[n2] << endl;
				
				bond newbond(dsb_atm[n1], dsb_atm[n2], bondtype('S'));
				tmpmdl->AddBond(newbond);
			}
		}
	}
	
	// secondary structure...
	
	// the default state is always loop. strands have the same places as in K&S.
	// helices are shifted: the smallest helical peptide is "4...." -> "LHHHL"!!!!!
	
	for (i32u n1 = 0;n1 < lookup.size();n1++)
	{
		if (!lookup[n1]) continue;
		
		if (ci_vector[n1].GetLength() != (i32s) lookup[n1]->size())
		{
			cout << "BOG: the sizes of chn_info and lookup differ!!!" << endl;
			exit(EXIT_FAILURE);
		}
		
		const char * state = ci_vector[n1].GetState();
		for (i32s n2 = 0;n2 < (i32s) lookup[n1]->size();n2++)
		{
			lookup[n1]->operator[](n2)->flags &= (~(ATOMFLAG_SF_HELIX | ATOMFLAG_SF_STRAND));
			
			switch (state[n2])
			{
				case 'S':
				lookup[n1]->operator[](n2)->flags |= ATOMFLAG_SF_STRAND;
				break;
			}
		}
				
		for (i32s n2 = 2;n2 < ((i32s) lookup[n1]->size()) - 2;n2++)
		{
			switch (state[n2 - 2])
			{
				case '4':
				lookup[n1]->operator[](n2 - 1)->flags &= (~(ATOMFLAG_SF_STRAND));
				lookup[n1]->operator[](n2 + 0)->flags &= (~(ATOMFLAG_SF_STRAND));
				lookup[n1]->operator[](n2 + 1)->flags &= (~(ATOMFLAG_SF_STRAND));
				lookup[n1]->operator[](n2 - 1)->flags |= ATOMFLAG_SF_HELIX;
				lookup[n1]->operator[](n2 + 0)->flags |= ATOMFLAG_SF_HELIX;
				lookup[n1]->operator[](n2 + 1)->flags |= ATOMFLAG_SF_HELIX;
				break;
			}
		}
	}
	
	// delete the intermediate stuff and leave...
	
	for (i32u n1 = 0;n1 < lookup.size();n1++)
	{
		if (lookup[n1] != NULL) delete lookup[n1];
	}
	
	return tmpmdl;
}

void model::aa2sf_Convert_sub2(iter_al * iter, vector<i32s> & idv, fGL * crd, i32u cset)
{
	vector<atom *> refv;
	for (i32u n1 = 0;n1 < idv.size();n1++)
	{
		iter_al it2 = iter[0];
		while (it2 != iter[1] && ((* it2).res_id & 0xFF) != idv[n1]) it2++;
		if (it2 != iter[1]) refv.push_back(& (* it2));
	}
	
	if (!refv.size()) { cout << "BUG: no atoms found!" << endl; exit(EXIT_FAILURE); }
	
	for (i32u n1 = 0;n1 < 3;n1++)
	{
		crd[n1] = 0.0;
		for (i32u n2 = 0;n2 < refv.size();n2++)
		{
			const fGL * cdata = refv[n2]->GetCRD(cset);
			crd[n1] += cdata[n1];
		}
		
		crd[n1] /= (f64) refv.size();
	}
}

/*##############################################*/
/*##############################################*/

// what comes to PDB chain id's here, we follow the "pdb-select principle" by converting
// the empty id ' ' with '_' to improve readability...

readpdb_mdata * model::readpdb_ReadMData(const char * filename)
{
	cout << "reading PDB metadata from file " << filename << endl;
	
	readpdb_mdata * mdata = new readpdb_mdata;
	
	char buffer[1024];
	
	ifstream file(filename, ios::in);
	if (file.fail())
	{
		cout << "file \"" << filename << "\" not found." << endl;
		
		file.close();
		return mdata;
	}
	
	while (!file.eof())
	{
		char record_id[8];
		for (i32s n1 = 0;n1 < 6;n1++)
		{
			char tchar = (char) file.get();
			if (tchar != ' ') record_id[n1] = tchar;
			else record_id[n1] = 0;
		}
		
		record_id[6] = 0;	// terminate the record string...
		
		if (!strcmp("SEQRES", record_id))
		{
			i32s blocknum; file >> blocknum;
			file.get(); char chn_id = (char) file.get();
			i32s chn_length; file >> chn_length;
			
			if (chn_id == ' ') chn_id = '_';	// fix the empty chain id...
			
			i32s chn_num;
			if (blocknum == 1)	// this is a new record...
			{
				cout << "found a new chain " << mdata->chn_vector.size();
				cout << " '" << chn_id << "' with " << chn_length << " residues." << endl;
				
				readpdb_mdata_chain * new_data = new readpdb_mdata_chain;
				new_data->chn_id = chn_id; new_data->seqres = new char[chn_length + 1];
				
				for (i32s n1 = 0;n1 < chn_length;n1++) new_data->seqres[n1] = '?';
				new_data->seqres[chn_length] = 0;	// terminate the string!!!
				
				chn_num = mdata->chn_vector.size();
				mdata->chn_vector.push_back(new_data);
			}
			else		// this is an old record, find it...
			{
				chn_num = 0;
				while (chn_num < ((i32s) mdata->chn_vector.size()))
				{
					if (mdata->chn_vector[chn_num]->chn_id == chn_id) break;
					else chn_num++;
				}
				
				if (chn_num == ((i32s) mdata->chn_vector.size()))
				{
					cout << "BUG : readpdb_ReadMData : unknown chain found!!!" << endl;
					exit(EXIT_FAILURE);
				}
			}
			
			chn_length = strlen(mdata->chn_vector[chn_num]->seqres);
			i32s counter = (blocknum - 1) * 13;
			
			for (i32s n1 = 0;n1 < 13;n1++)
			{
				char residue[16];
				file >> residue;
				
				char symbol = '?';
				if (!strcmp("ALA", residue)) symbol = 'A';
				if (!strcmp("ARG", residue)) symbol = 'R';
				if (!strcmp("ASN", residue)) symbol = 'N';
				if (!strcmp("ASP", residue)) symbol = 'D';
				if (!strcmp("CYS", residue)) symbol = 'C';
				if (!strcmp("GLN", residue)) symbol = 'Q';
				if (!strcmp("GLU", residue)) symbol = 'E';
				if (!strcmp("GLY", residue)) symbol = 'G';
				if (!strcmp("HIS", residue)) symbol = 'H';
				if (!strcmp("ILE", residue)) symbol = 'I';
				if (!strcmp("LEU", residue)) symbol = 'L';
				if (!strcmp("LYS", residue)) symbol = 'K';
				if (!strcmp("MET", residue)) symbol = 'M';
				if (!strcmp("PHE", residue)) symbol = 'F';
				if (!strcmp("PRO", residue)) symbol = 'P';
				if (!strcmp("SER", residue)) symbol = 'S';
				if (!strcmp("THR", residue)) symbol = 'T';
				if (!strcmp("TRP", residue)) symbol = 'W';
				if (!strcmp("TYR", residue)) symbol = 'Y';
				if (!strcmp("VAL", residue)) symbol = 'V';
				
				mdata->chn_vector[chn_num]->seqres[counter++] = symbol;
				if (counter == chn_length) break;
			}
		}
		
		file.getline(buffer, sizeof(buffer));		// move to the next line...
	}
	
	file.close();
	
	// ready...
	
	if (mdata->chn_vector.empty()) cout << "WARNING : no chains found!!!" << endl;
	cout << "done." << endl;
	
	return mdata;
}

// chn_num must be either NOT_DEFINED (which is -1 meaning all chains) or the chain index in mdata...
// chn_num must be either NOT_DEFINED (which is -1 meaning all chains) or the chain index in mdata...
// chn_num must be either NOT_DEFINED (which is -1 meaning all chains) or the chain index in mdata...

//#define READPDB_ENABLE_MULTIPLE_CRDSETS	// enable this to read all data from NMR entries...

void model::readpdb_ReadData(const char * filename, readpdb_mdata * mdata, i32s chn_num)
{
	cout << "reading PDB data from file " << filename << endl;
	
	ifstream file(filename, ios::in);
	if (file.fail())
	{
		cout << "file \"" << filename << "\" not found." << endl;
		
		file.close();
		return;
	}
	
	// read the whole file to temporary arrays...
	
	vector<readpdb_data_atom> atom_data;
	vector<readpdb_data_ssbond> ssbond_data;
	
	i32s model_counter = 0;
	i32s atom_counter = NOT_DEFINED;
	
	char buffer[1024];
	
	while (!file.eof())
	{
		char record_id[8];
		for (i32s n1 = 0;n1 < 6;n1++)
		{
			char tchar = (char) file.get();
			if (tchar != ' ') record_id[n1] = tchar;
			else record_id[n1] = 0;
		}
		
		record_id[6] = 0;	// terminate the record string...
		
		if (!strcmp("ATOM", record_id))
		{
			i32s serial_number; file >> serial_number;
			
			for (i32s n1 = 0;n1 < 18;n1++) buffer[n1] = (char) file.get();
			buffer[5] = ' '; buffer[15] = ' ';
			
			char chn_id; i32s res_num; char res_name[5]; char atm_name[5]; fGL crd[3];
			
			istrstream str(buffer, sizeof(buffer));
			str >> atm_name >> res_name; str.get(); chn_id = (char) str.get(); str >> res_num;
			file >> crd[0]; crd[0] /= 10.0; file >> crd[1]; crd[1] /= 10.0; file >> crd[2]; crd[2] /= 10.0;
			
			if (chn_id == ' ') chn_id = '_';	// fix the empty chain id...
			
			bool test1 = (chn_num == NOT_DEFINED);
			bool test2 = test1 ? false : (chn_id == mdata->chn_vector[chn_num]->chn_id);
			
			if (test1 || test2)
			{
				if (model_counter == 0)
				{
					readpdb_data_atom new_data;
					new_data.ref = NULL;
					
					strcpy(new_data.atm_name, atm_name);
					strcpy(new_data.res_name, res_name);
					
					new_data.chn_id = chn_id;
					new_data.res_num = res_num;
					
					new_data.crd[model_counter][0] = crd[0];
					new_data.crd[model_counter][1] = crd[1];
					new_data.crd[model_counter][2] = crd[2];
					
					atom_data.push_back(new_data);
				}
				else
				{
					bool test1 = !strcmp(atom_data[atom_counter].res_name, res_name);
					bool test2 = !strcmp(atom_data[atom_counter].atm_name, atm_name);
					
					if (!test1) { cout << "BUG: test1 failed!!!" << endl; exit(EXIT_FAILURE); }
					if (!test2) { cout << "BUG: test2 failed!!!" << endl; exit(EXIT_FAILURE); }
					
					atom_data[atom_counter].crd[model_counter][0] = crd[0];
					atom_data[atom_counter].crd[model_counter][1] = crd[1];
					atom_data[atom_counter].crd[model_counter][2] = crd[2];
					
					atom_counter++;
				}
			}
		}
		
		if (!strcmp("SSBOND", record_id))
		{
			i32s serial_number; file >> serial_number;
			readpdb_data_ssbond new_data; new_data.ref = NULL;
			
			for (i32s n1 = 0;n1 < 2;n1++)
			{
				file >> buffer; file.get();
				char chn_id = (char) file.get();
				
				if (chn_id == ' ') chn_id = '_';	// fix the empty chain id...
				
				new_data.chn_id = chn_id;
				
				for (i32s n2 = 0;n2 < 6;n2++) buffer[n2] = (char) file.get();
				istrstream str(buffer, sizeof(buffer));
				str >> new_data.res_num;
				
				ssbond_data.push_back(new_data);
			}
		}
		
		// how to deal with multiple coordinate sets (often present in NMR entries)???
		// how to deal with multiple coordinate sets (often present in NMR entries)???
		// how to deal with multiple coordinate sets (often present in NMR entries)???
		
#ifdef READPDB_ENABLE_MULTIPLE_CRDSETS

		// in this caseat the moment, we will read multiple records up to READPDB_MAX_CRDSETS.
		// if READPDB_MAX_CRDSETS is exceeded, the program stops; must recompile then...
		
		if (!strcmp("MODEL", record_id))
		{
			i32s model_number; file >> model_number;
			
			if (model_number > 1)
			{
				model_counter++;
				
				if (model_counter == READPDB_MAX_CRDSETS)	// this should never happen...
				{
					cout << "BUG: READPDB_MAX_CRDSETS exceeded!!!" << endl;
					exit(EXIT_FAILURE);
				}
				
				atom_counter = 0;
			}
		}
		
#else	// READPDB_ENABLE_MULTIPLE_CRDSETS

		// in this case, we just look for "ENDMDL" record, which says this coordinate set is
		// now ready. so, we read only the first coordinate set. the "MODEL"/"ENDMDL" records
		// are only after "SSBOND", so it should not have big effects...
		
		if (!strcmp("ENDMDL", record_id))
		{
			cout << "ENDMDL record found, skipping the rest of this file..." << endl;
			break;
		}
		
#endif	// READPDB_ENABLE_MULTIPLE_CRDSETS

		file.getline(buffer, sizeof(buffer));		// move to the next line...
	}
	
	file.close();
	
	if (atom_data.empty())
	{
		cout << "no atoms found !!!" << endl;
		return;
	}
		
	// we have now read all the crd-sets; check if we need more space for them...
	
	i32s csets = cs_vector.size(); i32s new_csets = ((model_counter + 1) - csets);
	cout << "there were " << csets << " old crd-sets, creating " << new_csets << " new..." << endl;
	
	PushCRDSets(new_csets);
	
	for (i32u n1 = 0;n1 < cs_vector.size();n1++) SetCRDSetVisible(n1, true);
	
	// count the chains and relate them to mdata, create new records to mdata if necessary...
	
	vector<i32s> chn_index;
	char tmp_chn_id; i32u tmp_index;
	
	tmp_chn_id = atom_data.front().chn_id;
	
	tmp_index = 0;
	while (tmp_index < mdata->chn_vector.size())
	{
		if (tmp_chn_id == mdata->chn_vector[tmp_index]->chn_id) break;
		else tmp_index++;
	}
	
	if (tmp_index == mdata->chn_vector.size())
	{
		readpdb_mdata_chain * unknown = new readpdb_mdata_chain;
		unknown->chn_id = tmp_chn_id; unknown->seqres = NULL;
		mdata->chn_vector.push_back(unknown);
	}
	
	chn_index.push_back(tmp_index);
	
	for (i32u n1 = 1;n1 < atom_data.size();n1++)
	{
		tmp_chn_id = atom_data[n1].chn_id;
		if (tmp_chn_id == mdata->chn_vector[chn_index.back()]->chn_id) continue;
		else
		{
			tmp_index = 0;
			while (tmp_index < mdata->chn_vector.size())
			{
				if (tmp_chn_id == mdata->chn_vector[tmp_index]->chn_id) break;
				else tmp_index++;
			}
			
			if (tmp_index == mdata->chn_vector.size())
			{
				readpdb_mdata_chain * unknown = new readpdb_mdata_chain;
				unknown->chn_id = tmp_chn_id; unknown->seqres = NULL;
				mdata->chn_vector.push_back(unknown);
			}
			
			chn_index.push_back(tmp_index);
		}
	}
	
	// now create the chains, checking validity of the residues...
	
	for (i32u n1 = 0;n1 < chn_index.size();n1++)
	{
		char current_chn_id = mdata->chn_vector[chn_index[n1]]->chn_id;
		
		i32s current_chn_length;
		if (mdata->chn_vector[chn_index[n1]]->seqres != NULL)
		{
			current_chn_length = strlen(mdata->chn_vector[chn_index[n1]]->seqres);
		}
		else
		{
			current_chn_length = NOT_DEFINED;
		}
		
		i32s previous_residue = 0;
		
		i32s range1[2];
		
		range1[0] = 0;
		while (atom_data[range1[0]].chn_id != current_chn_id) range1[0]++;
		
		range1[1] = range1[0];
		while (range1[1] < (i32s) atom_data.size() && atom_data[range1[1]].chn_id == current_chn_id) range1[1]++;
		
		vector<i32s> res_data;
		
		i32s range2[2];
		range2[0] = range1[0];
		while (range2[0] < range1[1])
		{
			i32s residue = atom_data[range2[0]].res_num;
			const char * res_name = atom_data[range2[0]].res_name;
			
			range2[1] = range2[0];
			while (range2[1] < (i32s) atom_data.size() && atom_data[range2[1]].res_num == residue) range2[1]++;
			
			bool standard = false;
			if (!strcmp("ALA", res_name)) standard = true;
			if (!strcmp("ARG", res_name)) standard = true;
			if (!strcmp("ASN", res_name)) standard = true;
			if (!strcmp("ASP", res_name)) standard = true;
			if (!strcmp("CYS", res_name)) standard = true;
			if (!strcmp("GLN", res_name)) standard = true;
			if (!strcmp("GLU", res_name)) standard = true;
			if (!strcmp("GLY", res_name)) standard = true;
			if (!strcmp("HIS", res_name)) standard = true;
			if (!strcmp("ILE", res_name)) standard = true;
			if (!strcmp("LEU", res_name)) standard = true;
			if (!strcmp("LYS", res_name)) standard = true;
			if (!strcmp("MET", res_name)) standard = true;
			if (!strcmp("PHE", res_name)) standard = true;
			if (!strcmp("PRO", res_name)) standard = true;
			if (!strcmp("SER", res_name)) standard = true;
			if (!strcmp("THR", res_name)) standard = true;
			if (!strcmp("TRP", res_name)) standard = true;
			if (!strcmp("TYR", res_name)) standard = true;
			if (!strcmp("VAL", res_name)) standard = true;
			
			bool skip = false;
			if (standard)
			{
				if (readpdb_ReadData_sub1(atom_data, range2, "N", true) == NOT_DEFINED) skip = true;
				if (readpdb_ReadData_sub1(atom_data, range2, "CA", true) == NOT_DEFINED) skip = true;
				if (readpdb_ReadData_sub1(atom_data, range2, "C", true) == NOT_DEFINED) skip = true;
				if (readpdb_ReadData_sub1(atom_data, range2, "O", true) == NOT_DEFINED) skip = true;
			}
			else
			{
				cout << "could not recognize this residue: " << res_name << endl;
				skip = true;
			}
			
			if (skip)
			{
				cout << "skipping broken residue " << residue << " " << res_name << endl;
				atom_data.erase(atom_data.begin() + range2[0], atom_data.begin() + range2[1]);
				range1[1] += (range2[0] - range2[1]);
			}
			else
			{
				// ok, there seems to be no big problems with this residue.
				// first we make some bookkeeping with the original sequence...
				
				if ((previous_residue + 1) != residue)
				{
					for (i32s n2 = (previous_residue + 1);n2 < residue;n2++)
					{
						// the variables use "pascal/fortran-style" counting, beginning 1,
						// but we will store the values in "c-style" way, beginning 0!!!
						
						mdata->chn_vector[chn_index[n1]]->missing_residues.push_back(n2 - 1);
					}
				}
				
				previous_residue = residue;
				bool alpha_carbon_is_found = false;
				
				// ...and then we just create the atoms.
				
				res_data.push_back(range2[0]);
				for (i32s n2 = range2[0];n2 < range2[1];n2++)
				{
					strcpy(buffer, atom_data[n2].atm_name); buffer[1] = 0;		// here we suppose
					element el = element(buffer);					// single-char elements!!!
					
					if (el.GetAtomicNumber() == 1) continue;
					atom newatom(el, NULL, cs_vector.size());
					
					for (i32u n3 = 0;n3 < cs_vector.size();n3++)
					{
						fGL x = atom_data[n2].crd[n3][0];
						fGL y = atom_data[n2].crd[n3][1];
						fGL z = atom_data[n2].crd[n3][2];
						newatom.SetCRD(n3, x, y, z);
					}
					
					AddAtom(newatom); atom_data[n2].ref = (& atom_list.back());
					
					// if this atom was an alpha-carbon, then add it to the list... but alternative
					// locations might cause trouble here; make sure that only one c-alpha is added!!!
					
					if (!strcmp(atom_data[n2].atm_name, "CA") && !alpha_carbon_is_found)
					{
						mdata->chn_vector[chn_index[n1]]->alpha_carbons.push_back(atom_data[n2].ref);
						alpha_carbon_is_found = true;
					}
				}
				
				readpdb_ReadData_sub2(atom_data, range2, "N", "CA", 'S');
				readpdb_ReadData_sub2(atom_data, range2, "CA", "C", 'S');
				readpdb_ReadData_sub2(atom_data, range2, "C", "O", 'D');
				
				if (!strcmp("ALA", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
				}
				
				if (!strcmp("ARG", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "CD", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CD", "NE", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "NE", "CZ", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CZ", "NH1", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CZ", "NH2", 'C');
				}
				
				if (!strcmp("ASN", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "OD1", 'D');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "ND2", 'S');
				}
				
				if (!strcmp("ASP", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "OD1", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "OD2", 'C');
				}
				
				if (!strcmp("CYS", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "SG", 'S');
				}
				
				if (!strcmp("GLN", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "CD", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CD", "OE1", 'D');
					readpdb_ReadData_sub2(atom_data, range2, "CD", "NE2", 'S');
				}
				
				if (!strcmp("GLU", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "CD", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CD", "OE1", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CD", "OE2", 'C');
				}
				
				// GLY -> we don't have to do anything in this case...
				// GLY -> we don't have to do anything in this case...
				// GLY -> we don't have to do anything in this case...
				
				if (!strcmp("HIS", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "ND1", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "CD2", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "ND1", "CE1", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CD2", "NE2", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CE1", "NE2", 'C');
				}
				
				if (!strcmp("ILE", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG1", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG2", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG1", "CD1", 'S');
				}
				
				if (!strcmp("LEU", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "CD1", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "CD2", 'S');
				}
				
				if (!strcmp("LYS", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "CD", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CD", "CE", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CE", "NZ", 'S');
				}
				
				if (!strcmp("MET", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "SD", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "SD", "CE", 'S');
				}
				
				if (!strcmp("PHE", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "CD1", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "CD2", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CD1", "CE1", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CD2", "CE2", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CE1", "CZ", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CE2", "CZ", 'C');
				}
				
				if (!strcmp("PRO", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "CD", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CD", "N", 'S');
				}
				
				if (!strcmp("SER", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "OG", 'S');
				}
				
				if (!strcmp("THR", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "OG1", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG2", 'S');
				}
				
				if (!strcmp("TRP", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "CD1", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "CD2", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CD1", "NE1", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CD2", "CE2", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CD2", "CE3", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "NE1", "CE2", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CE2", "CZ2", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CE3", "CZ3", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CZ2", "CH2", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CZ3", "CH2", 'C');
				}
				
				if (!strcmp("TYR", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "CD1", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CG", "CD2", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CD1", "CE1", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CD2", "CE2", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CE1", "CZ", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CE2", "CZ", 'C');
					readpdb_ReadData_sub2(atom_data, range2, "CZ", "OH", 'S');
				}
				
				if (!strcmp("VAL", res_name))
				{
					readpdb_ReadData_sub2(atom_data, range2, "CA", "CB", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG1", 'S');
					readpdb_ReadData_sub2(atom_data, range2, "CB", "CG2", 'S');
				}
				
				range2[0] = range2[1];
			}
		}
		
		// the end of sequence bookkeeping : check if there were missing n-terminal residues.
		
		if (previous_residue < current_chn_length)
		{
			for (i32s n2 = previous_residue;n2 < current_chn_length;n2++)
			{
				mdata->chn_vector[chn_index[n1]]->missing_residues.push_back(n2);
			}
		}
		
		if (!mdata->chn_vector[chn_index[n1]]->missing_residues.empty())
		{
			char out1 = mdata->chn_vector[chn_index[n1]]->chn_id;
			i32u out2 = mdata->chn_vector[chn_index[n1]]->missing_residues.size();
			
			cout << "at chain '" << out1 << "' there were " << out2 << " missing residues:" << endl;
			
			for (i32u n2 = 0;n2 < mdata->chn_vector[chn_index[n1]]->missing_residues.size();n2++)
			{
				cout << mdata->chn_vector[chn_index[n1]]->missing_residues[n2] << " ";
			}
			
			cout << endl;
		}
		
		// add connecting bonds between residues...
		
		res_data.push_back(range1[1]);
		for (i32s n2 = 0;n2 < ((i32s) res_data.size()) - 2;n2++)
		{
			i32s r1[2] = { res_data[n2],  res_data[n2 + 1] };
			i32s r2[2] = { res_data[n2 + 1],  res_data[n2 + 2] };
			
			i32s ind1 = readpdb_ReadData_sub1(atom_data, r1, "C", false);
			i32s ind2 = readpdb_ReadData_sub1(atom_data, r2, "N", false);
			
			if (ind1 == NOT_DEFINED || ind2 == NOT_DEFINED)
			{
				cout << "BUG: could not connect ..." << endl;
				exit(EXIT_FAILURE);
			}
			else
			{
				bondtype bt = bondtype('S');
				bond newbond(atom_data[ind1].ref, atom_data[ind2].ref, bt);
				AddBond(newbond);
			}
		}
		
		// check the c-terminal residue...
		
		if (res_data.size() < 2) continue;
		i32s last[2] = { res_data[res_data.size() - 2], res_data[res_data.size() - 1] };
		
		i32s ind1 = readpdb_ReadData_sub1(atom_data, last, "C", false);
		i32s ind2 = readpdb_ReadData_sub1(atom_data, last, "O", false);
		
		if (ind1 == NOT_DEFINED || ind2 == NOT_DEFINED)
		{
			cout << "BUG: could not find c-term group..." << endl;
			exit(EXIT_FAILURE);
		}
		else
		{
			bondtype bt = bondtype('C');
			bond tmpbond(atom_data[ind1].ref, atom_data[ind2].ref, bt);
			iter_bl it1 = find(bond_list.begin(), bond_list.end(), tmpbond);
			if (it1 != bond_list.end()) (* it1).bt = bt;
			
			ind2 = readpdb_ReadData_sub1(atom_data, last, "OXT", false);
			if (ind2 != NOT_DEFINED)
			{
				bond newbond(atom_data[ind1].ref, atom_data[ind2].ref, bt);
				AddBond(newbond);
			}
			else
			{
				cout << "missing terminal oxygen..." << endl;
				
				i32s ind[3];
				ind[0] = readpdb_ReadData_sub1(atom_data, last, "CA", false);
				ind[1] = readpdb_ReadData_sub1(atom_data, last, "C", false);
				ind[2] = readpdb_ReadData_sub1(atom_data, last, "O", false);
				
				const fGL * ref1; const fGL * ref2;
				
				ref1 = atom_data[ind[1]].ref->GetCRD(0);
				ref2 = atom_data[ind[0]].ref->GetCRD(0);
				v3d<fGL> v1 = v3d<fGL>(ref1, ref2);
				
				ref1 = atom_data[ind[1]].ref->GetCRD(0);
				ref2 = atom_data[ind[2]].ref->GetCRD(0);
				v3d<fGL> v2 = v3d<fGL>(ref1, ref2);
				
				fGL tmp2 = v1.len(); fGL tmp3 = v1.spr(v2) / (tmp2 * tmp2);
				v1 = v1 * tmp3; v2 = v2 - v1;
				
				fGL tmp4[3];
				const fGL * cdata = atom_data[ind[1]].ref->GetCRD(0);
				for (i32s n1 = 0;n1 < 3;n1++)
				{
					tmp4[n1] = cdata[n1];
					tmp4[n1] += v1.data[n1] - v2.data[n1];
				}
				
				element el = element(8);
				atom newatom(el, tmp4, cs_vector.size());
				AddAtom(newatom);
				
				bond newbond(atom_data[ind1].ref, (& atom_list.back()), bt);
				AddBond(newbond);
			}
		}
	}
	
	// add disulphide bonds...
	
	for (i32u n1 = 0;n1 < ssbond_data.size();n1++)
	{
		i32u counter = 0;
		while (counter < atom_data.size())
		{
			bool test1 = (ssbond_data[n1].chn_id == atom_data[counter].chn_id);
			bool test2 = (ssbond_data[n1].res_num == atom_data[counter].res_num);
			bool test3 = !strcmp(atom_data[counter].atm_name, "SG");
			if (test1 && test2 && test3) break; else counter++;
		}
		
		if (counter == atom_data.size()) continue;
		ssbond_data[n1].ref = atom_data[counter].ref;
	}
	
	for (i32u n1 = 0;n1 < ssbond_data.size();n1 += 2)
	{
		if (ssbond_data[n1 + 0].ref == NULL || ssbond_data[n1 + 1].ref == NULL)
		{
			cout << "could not create bridge ";
			cout << ssbond_data[n1 + 0].chn_id << " " << ssbond_data[n1 + 0].res_num << " -> ";
			cout << ssbond_data[n1 + 1].chn_id << " " << ssbond_data[n1 + 1].res_num << endl;
		}
		else
		{
			bondtype bt = bondtype('S');
			bond newbond(ssbond_data[n1 + 0].ref, ssbond_data[n1 + 1].ref, bt);
			AddBond(newbond);
		}
	}
	
	// ready...
	
	cout << "done." << endl;
	
	CenterCRDSet(0);
}

i32s model::readpdb_ReadData_sub1(vector<readpdb_data_atom> & adata, i32s * range, const char * atom_name, bool flag)
{
	for (i32s n1 = range[0];n1 < range[1];n1++)
	{
		if (!strcmp(adata[n1].atm_name, atom_name)) return n1;
	}
	
	cout << "atom " << atom_name << " is missing..." << endl;
	return NOT_DEFINED;
}

void model::readpdb_ReadData_sub2(vector<readpdb_data_atom> & adata, i32s * range, const char * at1, const char * at2, char btype)
{
	i32s ind1 = readpdb_ReadData_sub1(adata, range, at1, false);
	i32s ind2 = readpdb_ReadData_sub1(adata, range, at2, false);
	
	if (ind1 == NOT_DEFINED || ind2 == NOT_DEFINED) return;
	
	bondtype bt = bondtype(btype);
	bond newbond(adata[ind1].ref, adata[ind2].ref, bt);
	AddBond(newbond);
}

/*################################################################################################*/

crd_set::crd_set(void)
{
	description = NULL;
	
	accum_weight = 1.0;		// accum_value is not updated here?!?!?!?!?!
	visible = false;
}

crd_set::crd_set(const crd_set & p1)
{
	description = NULL;
	SetDescription(p1.description);
	
	accum_weight = p1.accum_weight;
	accum_value = p1.accum_value;
	visible = p1.visible;
}

crd_set::~crd_set(void)
{
	if (description != NULL) delete[] description;
}

void crd_set::SetDescription(const char * p1)
{
	if (description != NULL) delete[] description;
	
	if (p1 != NULL)
	{
		description = new char[strlen(p1) + 1];
		strcpy(description, p1);
	}
	else description = NULL;
}

/*################################################################################################*/

// eof
