// MODEL.H : the "model" class that stores the "atom" and "bond" objects.

// Copyright (C) 1998, 2002 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#include "config.h"	// this is target-dependent...

#ifndef MODEL_H
#define MODEL_H

class model;
class crd_set;

// the readpdb classes here are not fundamental, but are located here for convenience...
// the readpdb classes here are not fundamental, but are located here for convenience...
// the readpdb classes here are not fundamental, but are located here for convenience...

class readpdb_mdata;
struct readpdb_mdata_chain;
struct readpdb_data_atom;
struct readpdb_data_ssbond;

/*################################################################################################*/

class setup;		// submodel.h
class engine;		// engine.h

class geomopt_param;	// geomopt.h
class moldyn_param;	// moldyn.h

#include "factory.h"
#include "seqbuild.h"

#include <list>
#include <vector>
#include <fstream>
#include <iostream>
#include <algorithm>
using namespace std;

#define NOT_FOUND 0x7fffffff   // numeric_limits<i32s>::max()?!?!?!

#define SF_NUM_TYPES 37

/*################################################################################################*/

/// The model class contains information about all atoms and bonds in a model.

class model
{
	private:
	
	class_factory * factory;
	
	static char local_fn_buffer[256];

	protected:
	
	setup * current_setup;		///< Must always be a valid pointer!!!
	
	list<atom> atom_list;
	list<bond> bond_list;
	
	vector<crd_set *> cs_vector;	///< This determines how many crd_sets there are in the model.
	i32u crd_table_size_glob;	///< This determines how big the crd_table arrays are; always >= than above!!!
	
	bool is_index_clean;		///< Some flags that show which information is up-to-date...
	bool is_groups_clean;		///< Some flags that show which information is up-to-date...
	
	i32s nmol;
	vector<chn_info> * ref_civ;	// vector<chn_info *> ?!?!?!
	
	ifstream * trajfile;
	i32s total_traj_frames;
	i32s current_traj_frame;
	
// For thermodynamics stuff several "topologies" (and therefore also several elements,
// atom types etc...) are needed. COMMON topology???? will affect typerules!!! NOT REALLY USED YET!
//	i32s ctop;		// current topology (0 or 1) for td?????
//	bool dual_topology;

	public:

	i32s qm_total_charge;
	i32s qm_current_orbital;
	
	public:
	//protected:	// the dialog boxes need these...
	
	bool periodic;
	f64 box_hdim[3];
	f64 box_fdim[3];

	public:
	
	static const char sf_symbols[20 + 1];
	static const i32s sf_types[SF_NUM_TYPES];
	
	err_util * err;		///< Error reporting tools...
	prefs * model_prefs;	///< Preferences tools...
	
	public:
	
	static sequencebuilder amino_builder;
	static sequencebuilder nucleic_builder;
		
	friend class geomopt;
	friend class moldyn;
	
	friend class ribbon;
	
	friend class sequencebuilder;
	
	friend class file_trans;
	
	friend class setup_druid;	// needs access to current_setup...
	
	friend void CopyCRD(model *, engine *, i32u);
	friend void CopyCRD(engine *, model *, i32u);
	
	friend fGL plot_GetESPValue(fGL *, model *, fGL *);
	friend fGL plot_GetVDWSValue(fGL *, model *, fGL *);
	
	friend void DefineSecondaryStructure(model *);
	friend f64 HBondEnergy(model *, i32s *, i32s *);
	
	public:
	
	model(class_factory &);
	virtual ~model(void);
	
	class_factory * GetFactory(void) { return factory; }
	setup * GetCurrentSetup(void) { return current_setup; }
	
	const char * GetProjectFileNameExtension(void);
	
	static const char * GetFullPrimaryFN(const char *);
	static const char * GetFullSecondaryFN(const char *, const char *);
	
	static void OpenParameterFile(ifstream &, bool, const char *);
	
	void UpdateAccumValues(void);
	
/// The project-class will override this function...
	virtual void UpdateAllGraphicsViews(bool = false) { }
	
/// Add a message string to the logfile. This is just a default for TARGET2...
	virtual void PrintToLog(const char * p1) { cout << "PrintToLog: " << p1 << endl; }
	
/// This will return the number of coordinate sets.
/** It is supposed that at least one coordinate set exists all the time!!! */
	i32u GetCRDSetCount(void);
	
	bool GetCRDSetVisible(i32u);
	void SetCRDSetVisible(i32u, bool);
	
	void PushCRDSets(i32u);
	void PopCRDSets(i32u);
	
	void CopyCRDSet(i32u, i32u);
	void SwapCRDSets(i32u, i32u);
	
	void CenterCRDSet(i32u);
	void ReserveCRDSets(i32u);

	virtual void DiscardCurrEng(void);
	void SetupPlotting(void);
	
	// methods for adding new atoms and bonds:
	// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	
	void AddAtom(atom &);		///< This will just push new atom to the atom list.
	void RemoveAtom(iter_al);	///< This will delete all bonds associated with this atom, and erase atom from the list...
	
	void AddBond(bond &);		///< This will add neighbor infos for both atoms and add the new bond into the bond list.
	void RemoveBond(iter_bl);	///< This will remove infos from the atoms and erase bond from the bond list.
	
	void Clear(void);		///< This will remove all atoms and bonds.
	
	// methods for accessing atom/bond lists:
	// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	
	iter_al GetAtomsBegin(void) { return atom_list.begin(); }	//const;
	iter_al GetAtomsEnd(void) { return atom_list.end(); }		//const;
	
	iter_bl GetBondsBegin(void) { return bond_list.begin(); }	//const;
	iter_bl GetBondsEnd(void) { return bond_list.end(); }		//const;
	
	// methods for file I/O...
	// ^^^^^^^^^^^^^^^^^^^^^^^
	
	bool ReadStream_OLD(istream &, bool, bool = false);		///< this is for the very old version.
	bool ReadStream_v100(istream &, bool, bool = false);		///< this is for the version 1.00.
//	bool ReadStream_v110(istream &, bool, bool = false);		// this will be for the version 1.10, etc...
	
/// This is an input function for the v1.10 ghemical file format.
	bool ReadStream(istream &, bool, bool = false);		// virtual
	
/// This is an output function for the v1.10 ghemical file format.
	void WriteStream(ostream &);				// virtual
	
	bool ImportFile(const char *filename, int index = 0);	// virtual

	// add or remove hydrogen atoms...
	// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	
	void AddHydrogens(void);
	void RemoveHydrogens(void);
	
	i32s GetQMTotalCharge(void) { return qm_total_charge; }
	i32s GetQMCurrentOrbital(void) { return qm_current_orbital; }
	
	i32s GetAtomCount(void) { return atom_list.size(); }	///< This will return atom_list.size().
	i32s GetBondCount(void) { return bond_list.size(); }	///< This will return bond_list.size().
	
	i32s GetMoleculeCount(void) { return nmol; }			///< This will return nmol (see UpdateGroups() for more info).

	bool IsEmpty(void) { return (bond_list.empty() && atom_list.empty()); }		///< This will return whether the system is empty.
	
	bool IsIndexClean(void) { return is_index_clean; }
	bool IsGroupsClean(void) { return is_groups_clean; }
	
	iter_al FindAtomByIndex(i32s);
	
	void GetRange(i32s, i32s, iter_al *);	///< This is just a default version of GetRange() using the full range of atom list iterators...
	
/// GetRange is used to get a range of atoms that form molecules, residues etc...
/** This will reduce the initial range of two atom list iterators to some subrange
with certain values in certain atom::id[]-fields. Before using this you MUST call 
model::UpdateGroups() to arrange the atom list! What the above explanation really tries 
to say, is that using this function you can pick up a certain part of the model; for 
example a molecule, or a chain in a macromolecule, or a range of residue in a chain. */
	void GetRange(i32s, iter_al *, i32s, iter_al *);
	
	i32s FindPath(atom *, atom *, i32s, i32s, i32s = 0);
	bool FindRing(atom *, atom *, signed char *, i32s, i32s, i32s = 0);
	
/** Adding or removing atoms or bonds will generally scramble the grouping, 
and when this happens one should call this function to discard all grouping information. */
	void InvalidateGroups(void);
	
	void UpdateIndex(void);
	
/** This will group the atom list so that molecules (and chains/residues if defined) 
will form continuous groups. Will validate model::nmol and permit the use of model::GetRange()-functions. */
	void UpdateGroups(void);

/** This will validate ref_civ and fill it with chain descriptions using all 
sequencebuilder objects. Finally it will sort the atom list so that chains and residues 
will be ordered sequentially in contiguous blocks. */
	void UpdateChains(void);
	
	private:
	
/** This will set molecule numbers quickly using a recursive search algorithm. 
This is private because only model::UpdateGroups() should use this... */
	void GatherAtoms(atom *, i32s);
	
	public:
	
	// here we have a set of Do???()-functions. the idea is that there is a set
	// of features that we wish to behave EXACTLY same way in each target/platform.
	// we then create a Do???()-function for the feature, and hide the details of
	// the user interface in a set of virtual functions.
	
/** This will perform an energy calculation, and report the result. */
	void DoEnergy(void);
	
/** This will perform geometry optimization. */
	void DoGeomOpt(geomopt_param &);
	
/** This is used to ask user the GO options; this default function will just silently accept the defaults. */
	virtual void GeomOptGetParam(geomopt_param &);
	
/** This will perform molecular dynamics. */
	void DoMolDyn(moldyn_param &);
	
/** This is used to ask user the MD options; this default function will just silently accept the defaults. */
	virtual void MolDynGetParam(moldyn_param &);
	
	void DoRandomSearch(i32s, i32s);		///< This will perform a random search using torsions as variables.
	void DoSystematicSearch(i32s, i32s);
	void DoMonteCarloSearch(i32s, i32s, i32s);
	
/** This will print the molecular formula and weight. */
	void DoFormula(void);
	
	// methods related to MD trajectories...
	// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	
	void OpenTrajectory(const char *);
	void CloseTrajectory(void);
	void ReadFrame(void);
	// WriteFrame???
	
	i32s GetTotalFrames(void);
	ifstream * GetTrajectoryFile(void);
	
	i32s GetCurrentFrame(void);
	void SetCurrentFrame(i32s);
	
	public:
	
/** This is used to update the atom properties that are not saved into files (masses, radii etc) for sf atoms. */
	void sf_UpdateAtomProperties(void);
	
	// these functions are for all-atoms / simplified conversions...
	// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	
	void aa2sf_ConvertAll(readpdb_mdata *);
	void aa2sf_ConvertSelection(readpdb_mdata *);
	model * aa2sf_Convert_sub1(readpdb_mdata *, bool);
	void aa2sf_Convert_sub2(iter_al *, vector<i32s> &, fGL *, i32u);
	
/// A function for reading in Brookhaven PDB/ENT files.
/** The readpdb functions here are used to import PDB files as correctly as possible. 
The PDB files represent experimental results, and in many cases the structures in files 
have gapped and/or incomplete sequences, incomplete residues, and so on.

The readpdb functions do the import in two stages: in first stage read in the "metadata" 
(all headers and remarks about the data including the original sequence), and in second stage 
read in the data as correctly as possible. later, results from these two can be compared, for 
example to evaluate quality of the data or to match the data with records in other databases. */
	readpdb_mdata * readpdb_ReadMData(const char *);
	
	void readpdb_ReadData(const char *, readpdb_mdata *, i32s);
	i32s readpdb_ReadData_sub1(vector<readpdb_data_atom> &, i32s *, const char *, bool);
	void readpdb_ReadData_sub2(vector<readpdb_data_atom> &, i32s *, const char *, const char *, char);
};

/*################################################################################################*/

/// A coordinate-set class.

class crd_set
{
	protected:
	
	char * description;
	
// bring also the coloring information here??? -> would allow different colors for different crdsets!!!
// bring also the coloring information here??? -> would allow different colors for different crdsets!!!
// bring also the coloring information here??? -> would allow different colors for different crdsets!!!

	public:
	
// these are tricky to protect, since these are still developing and are not yet used much.
// so, these are public as long as some reasonable strategy is found...

	fGL accum_weight;
	fGL accum_value;
	bool visible;
	
	public:
	
	crd_set(void);
	crd_set(const crd_set &);
	~crd_set(void);
	
	void SetDescription(const char *);
};

/*################################################################################################*/

// define struct readpdb_mdata_chain before class readpdb_mdata, since the latter uses
// former in some inline functions...

// how to best relate readpdb_mdata_chain and chn_info !??!?!?!?!?
// maybe just by storing the alpha-carbon pointers here...

struct readpdb_mdata_chain
{
	char chn_id;
	char * seqres;
	
	vector<i32s> missing_residues;
	vector<atom *> alpha_carbons;
};

// class readpdb_mdata is a class just to make the memory management easier. the data members in
// the class are filled in model::readpdb_ReadMData(), and at end the object is just to be deleted.

class readpdb_mdata
{
	public:
	
	vector<readpdb_mdata_chain *> chn_vector;
	
	public:
	
	readpdb_mdata(void)
	{
	}
	
	~readpdb_mdata(void)
	{
		for (i32u n1 = 0;n1 < chn_vector.size();n1++)
		{
			delete[] chn_vector[n1]->seqres;	// delete the sequence...
			delete chn_vector[n1];			// delete the whole record...
		}
	}
};

// READPDB_MAX_CRDSETS is relevant only if READPDB_ENABLE_MULTIPLE_CRDSETS is defined...
// READPDB_MAX_CRDSETS is relevant only if READPDB_ENABLE_MULTIPLE_CRDSETS is defined...
// READPDB_MAX_CRDSETS is relevant only if READPDB_ENABLE_MULTIPLE_CRDSETS is defined...

#define READPDB_MAX_CRDSETS 10

struct readpdb_data_atom
{
	char chn_id;
	
	i32s res_num;
	char res_name[5];
	char atm_name[5];
	fGL crd[READPDB_MAX_CRDSETS][3];
	
	atom * ref;
};

struct readpdb_data_ssbond
{
	char chn_id;
	i32s res_num;
	
	atom * ref;
};

/*################################################################################################*/

#endif	// MODEL_H

// eof
