// PROJECT.H : a "document/view"-style base class for all graphical user interfaces.

// Copyright (C) 1998, 2002 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#include "config.h"	// this is target-dependent...

#ifndef PROJECT_H
#define PROJECT_H

#ifdef ENABLE_GRAPHICS

class project;

class color_mode;
class color_mode_element;		// atom *
class color_mode_state;			// atom *

class transparent_primitive;
class transparent_primitive_data;

class tpd_tri_3c;
class tpd_quad_4c;

/*################################################################################################*/

class dummy_object;		// objects.h
class smart_object;		// objects.h

class camera;			// camera.h
class light;			// camera.h

class transformer;		// camera.h

#include "model.h"
#include "views.h"

#include <vector>
using namespace std;

/*################################################################################################*/

#define SB_SIZE 512		// the OGL selection buffer size.

/**	The "##project" class is an enhanced version of "##model" class, containing
	graphical user interface.
*/

class project : public model
{
	private:
	
	graphics_class_factory * graphics_factory;
	
	protected:
	
	char * project_path;		// may or may not exist...
	char * project_filename;	// should always exist...
	
	vector<camera *> camera_vector;
	vector<graphics_view *> graphics_view_vector;
	vector<plotting_view *> plotting_view_vector;
	
/**	A static counter of display list IDs.
	In some targets there can be several GUI's running simultaneously,
	so we must make sure that the display list ID's are unique -> static.
*/
	static iGLu list_counter;
	
	vector<transparent_primitive> tp_vector;
	
#ifdef ENABLE_TREELIST_VIEW

	// at the moment, we have only a single treelist_view object.
	// this could be always open?!?!? is it needed??? improve MDI first...
	
	treelist_view * tlv;
	
#endif	// ENABLE_TREELIST_VIEW

	public:
	
	enum rmode { Normal = 0, Transform1 = 1, Transform2 = 2 };
	
	vector<light *> light_vector;
	vector<smart_object *> object_vector;
	dummy_object * selected_object;
	
	iGLu * select_buffer;

	public:
	
	static color_mode_element cm_element;
	static color_mode_state cm_state;
	
	atom * draw_data[2];
	
	public:
	
	project(graphics_class_factory &);
	virtual ~project(void);
	
	graphics_class_factory * GetGraphicsFactory(void) { return graphics_factory; }
	
/**	Some functions to support transparent primitives... 
	External objects (planes, surfaces, etc) should use these.
*/
	bool AddTP(void *, transparent_primitive &);
	void UpdateMPsForAllTPs(void *);
	void RemoveAllTPs(void *);
	
	// some file name functions...
	// ^^^^^^^^^^^^^^^^^^^^^^^^^^^
	
	void SetProjectPath(const char *);
	void SetProjectFileName(const char *);
	
	void SetDefaultProjectFileName(void);
	
	void ParseProjectFileNameAndPath(const char *);
	
	void GetProjectFileName(char *, int, bool);
	void GetFullProjectFileName(char *, int);
	
	// some utility functions.
	// ^^^^^^^^^^^^^^^^^^^^^^^

	const char * GetType(void);
	color_mode * GetDefaultColorMode(void);
	
	void SelectAll(void);
	void InvertSelection(void);

	// how output could co-operate in different targets??? these event_???() try that...
	// how output could co-operate in different targets??? these event_???() try that...
	// how output could co-operate in different targets??? these event_???() try that...
	
	virtual void event_SelectedObjectChanged(void) = 0;
	
	iGLu GetDisplayListIDs(iGLu);
	void DeleteDisplayLists(iGLu, iGLu);
	
	// some light-related functions:
	// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	
	i32s IsLight(const dummy_object *);
	bool SelectLight(const dummy_object *);
	
	bool AddGlobalLight(light *);
	bool AddLocalLight(light *, camera *);
	bool RemoveLight(dummy_object *);
	
	iGLs CountGlobalLights(void);
	iGLs CountLocalLights(camera *);
	
	bool SetupLights(camera *);
	
	void SetGlobalLightNumbers(void);
	void SetLocalLightNumbers(camera *);
	
	// some object-related functions:
	// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	
	i32s IsObject(const dummy_object *);
	bool SelectObject(const dummy_object *);
	
	void AddObject(smart_object *);
	bool RemoveObject(dummy_object *);
	
	// some camera/window functions:
	// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	
	fGL GetDefaultFocus(void);
	
	graphics_view * AddGraphicsView(bool = false);			// using a new camera...
	graphics_view * AddGraphicsView(camera *, bool = false);	// using an existing camera...
	
	bool IsThisLastGraphicsView(graphics_view *);
	bool RemoveGraphicsView(graphics_view *, bool = false);
	
#ifdef ENABLE_TREELIST_VIEW

	treelist_view * CreateTreeListView(void);
	bool DestroyTreeListView(void);
	
#endif	// ENABLE_TREELIST_VIEW

	enlevdiag_view * AddEnLevDiagView(void);

	void UpdateAllViews(void);
	virtual void UpdateAllWindowTitles(void);
	
	void UpdateAllGraphicsViews(bool = false);			// virtual
	void UpdateGraphicsViews(camera *, bool = false);
	void UpdateGraphicsView(graphics_view *, bool = false);
	
	void UpdatePlottingViews(void);
	
	// update for other views?!?!?!? or all views ?!?!?!?
	// update for other views?!?!?!? or all views ?!?!?!?
	// update for other views?!?!?!? or all views ?!?!?!?
	
/**	Contains some OpenGL initialization commands 
	(color, material, light settings). 
	The OGL selection buffer should also be assigned here!!!
*/
	void InitGL(void);
	
/**	Contains the OpenGL rendering commands that draw the 3D-view.

	Selection is done in a rather crude way by dumping the atom pointers directly to 
	the OGL selection buffer and reading them later there... Works fine as long as 
	both datatypes have the same size (currently it's 32 bits). If we some day need 
	64-bit pointers and have only 32-bit sel-buffer we have to use two names...
*/
	void Render(graphics_view *, rmode);

/**	This is just quickly cut into a different function for clarity... We will call this
	multiple times with translation applied if we render some periodic systems...
*/
	void RenderScene(graphics_view *, rmode, bool);
	
	void RenderAllTPs(graphics_view *, rmode);
	
/**	This should center the transformer object to the center of all
	selected atoms, and shift the center of all selected atoms to the origo.
*/
	void Center(transformer *);
	
/**	This should get the transformation matrix from the transformer
	object and use it to transform all selected atoms/bonds/whatever.
*/
	void Transform(transformer *);
	
	void DrawEvent(graphics_view *, vector<iGLu> &);
	void EraseEvent(graphics_view *, vector<iGLu> &);
	void SelectEvent(graphics_view *, vector<iGLu> &);
	void MeasureEvent(graphics_view *, vector<iGLu> &);
	
	// some common operations are implemented here:
	// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

	void ProcessCommandString(graphics_view *, const char *);
	
	void DoDeleteCurrentObject(void);
	
	void DoSwitchLocalLights(camera *, bool);
	void DoSwitchGlobalLights(camera *, bool);
	
	bool TestAtom(atom *, rmode);
	bool TestBond(bond *, rmode);
	
	void SetColor(color_mode *, atom *);
	
// here are some Do???() functions similar to those in the model class.
// however, the classes here are "interactive" and are directly dependent on graphics.
// so you don't want to use any of these classes in the console version...
	
	void DoEnergyPlot1D(i32s, i32s, i32s, i32s, i32s, fGL, fGL, i32s);
	void DoEnergyPlot2D(i32s, i32s, i32s, i32s, i32s, fGL, fGL, i32s, i32s, i32s, i32s, i32s, fGL, fGL, i32s);
};

float measure_len(float *, float *);
float measure_ang(float *, float *, float *);
float measure_tor(float *, float *, float *, float *);

/*################################################################################################*/

/// The "##color_mode" class is used to handle the color detemination details.

class color_mode
{
	protected:
	
// this is just an interface definition -> there is no relevant common data?!?!?
// this is just an interface definition -> there is no relevant common data?!?!?
// this is just an interface definition -> there is no relevant common data?!?!?

	public:
	
	color_mode(void) { }
	virtual ~color_mode(void) { }
	
// is this (void *)-method really the only possibility to give a virtual function different
// parameters without the "hiding problem"???????????????????

// should be [void *, i32s CRD_SET, fGL_a4] -> crd-set can be included in coloring??? ALSO NEED TO FIND A BETTER WAY TO INCLUDE PREFS!!!
// should be [void *, i32s CRD_SET, fGL_a4] -> crd-set can be included in coloring??? ALSO NEED TO FIND A BETTER WAY TO INCLUDE PREFS!!!
// should be [void *, i32s CRD_SET, fGL_a4] -> crd-set can be included in coloring??? ALSO NEED TO FIND A BETTER WAY TO INCLUDE PREFS!!!

	virtual void GetColor4(const void *, i32s, fGL *) = 0;
};

class color_mode_element : public color_mode
{
	protected:
	
	public:
	
	color_mode_element(void) : color_mode() { }
	~color_mode_element(void) { }
	
	void GetColor4(const void *, i32s, fGL *);
};

class color_mode_state : public color_mode
{
	protected:
	
	public:
	
	color_mode_state(void) : color_mode() { }
	~color_mode_state(void) { }
	
	void GetColor4(const void *, i32s, fGL *);
};

/*################################################################################################*/

/// An utility class for transparency effects...

class transparent_primitive
{
	private:
	
	fGL z_distance; void * owner;
	transparent_primitive_data * data;
	
	public:
	
	transparent_primitive(void);
	transparent_primitive(void *, transparent_primitive_data &);
	transparent_primitive(const transparent_primitive &);
	~transparent_primitive(void);
	
	bool TestOwner(void *) const;
	transparent_primitive_data * GetData(void) const;
	
	void UpdateDistance(const fGL *, const fGL *);
	
	bool operator<(const transparent_primitive &) const;
};

class transparent_primitive_data
{
	protected:
	
	fGL midpoint[3];
	
	public:
	
	transparent_primitive_data(void);
	virtual ~transparent_primitive_data(void);
	
	virtual void Render(void) = 0;
	virtual void UpdateMP(void) = 0;
	
	friend class transparent_primitive;
};

/// A triangle with 3 colors.

class tpd_tri_3c : public transparent_primitive_data
{
	protected:
	
	fGL * color[3];
	
	fGL * point[3];
	
	public:
	
	tpd_tri_3c(fGL *, fGL *, fGL *, fGL *, fGL *, fGL *);
	~tpd_tri_3c(void);
	
	void Render(void);		// virtual
	void UpdateMP(void);		// virtual
};

/// A quadrilateral with 4 colors.

class tpd_quad_4c : public transparent_primitive_data
{
	protected:
	
	fGL * color[4];
	
	fGL * point[4];
	
	public:
	
	tpd_quad_4c(fGL *, fGL *, fGL *, fGL *, fGL *, fGL *, fGL *, fGL *);
	~tpd_quad_4c(void);
	
	void Render(void);		// virtual
	void UpdateMP(void);		// virtual
};

/*################################################################################################*/

#endif	// ENABLE_GRAPHICS
#endif	// PROJECT_H

// eof
